"""Utilities for resolving types and forward references."""
import collections.abc
import copy
import dataclasses
import inspect
import sys
import types
import warnings
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    FrozenSet,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import Annotated, Self, get_args, get_origin, get_type_hints

from . import _fields, _unsafe_cache
from ._typing import TypeForm

TypeOrCallable = TypeVar("TypeOrCallable", TypeForm[Any], Callable)


def unwrap_origin_strip_extras(typ: TypeOrCallable) -> TypeOrCallable:
    """Returns the origin, ignoring typing.Annotated, of typ if it exists. Otherwise,
    returns typ."""
    # TODO: Annotated[] handling should be revisited...
    typ, _ = unwrap_annotated(typ)
    origin = get_origin(typ)

    if origin is not None:
        typ = origin

    return typ


def is_dataclass(cls: Union[TypeForm, Callable]) -> bool:
    """Same as `dataclasses.is_dataclass`, but also handles generic aliases."""
    return dataclasses.is_dataclass(unwrap_origin_strip_extras(cls))  # type: ignore


def resolve_generic_types(
    cls: TypeOrCallable,
) -> Tuple[TypeOrCallable, Dict[TypeVar, TypeForm[Any]]]:
    """If the input is a class: no-op. If it's a generic alias: returns the origin
    class, and a mapping from typevars to concrete types."""

    annotations: Tuple[Any, ...] = ()
    if get_origin(cls) is Annotated:
        # ^We need this `if` statement for an obscure edge case: when `cls` is a
        # function with `__tyro_markers__` set, we don't want/need to return
        # Annotated[func, markers].
        cls, annotations = unwrap_annotated(cls)

    # We'll ignore NewType when getting the origin + args for generics.
    origin_cls = get_origin(unwrap_newtype(cls)[0])
    type_from_typevar: Dict[TypeVar, TypeForm[Any]] = {}

    # Support typing.Self.
    # We'll do this by pretending that `Self` is a TypeVar...
    if hasattr(cls, "__self__"):
        self_type = getattr(cls, "__self__")
        if inspect.isclass(self_type):
            type_from_typevar[cast(TypeVar, Self)] = self_type
        else:
            type_from_typevar[cast(TypeVar, Self)] = self_type.__class__

    if (
        # Apply some heuristics for generic types. Should revisit this.
        origin_cls is not None
        and hasattr(origin_cls, "__parameters__")
        and hasattr(origin_cls.__parameters__, "__len__")
    ):
        typevars = origin_cls.__parameters__
        typevar_values = get_args(unwrap_newtype(cls)[0])
        assert len(typevars) == len(typevar_values)
        cls = origin_cls
        type_from_typevar.update(dict(zip(typevars, typevar_values)))

    if hasattr(cls, "__orig_bases__"):
        bases = getattr(cls, "__orig_bases__")
        for base in bases:
            origin_base = unwrap_origin_strip_extras(base)
            if origin_base is base or not hasattr(origin_base, "__parameters__"):
                continue
            typevars = origin_base.__parameters__
            typevar_values = get_args(base)
            type_from_typevar.update(dict(zip(typevars, typevar_values)))

    if len(annotations) == 0:
        return cls, type_from_typevar
    else:
        return (
            Annotated.__class_getitem__((cls, *annotations)),  # type: ignore
            type_from_typevar,
        )


@_unsafe_cache.unsafe_cache(maxsize=1024)
def resolved_fields(cls: TypeForm) -> List[dataclasses.Field]:
    """Similar to dataclasses.fields(), but includes dataclasses.InitVar types and
    resolves forward references."""

    assert dataclasses.is_dataclass(cls)
    fields = []
    annotations = get_type_hints(cast(Callable, cls), include_extras=True)
    for field in getattr(cls, "__dataclass_fields__").values():
        # Avoid mutating original field.
        field = copy.copy(field)

        # Resolve forward references.
        field.type = annotations[field.name]

        # Skip ClassVars.
        if get_origin(field.type) is ClassVar:
            continue

        # Unwrap InitVar types.
        if isinstance(field.type, dataclasses.InitVar):
            field.type = field.type.type

        fields.append(field)

    return fields


def is_namedtuple(cls: TypeForm) -> bool:
    return (
        hasattr(cls, "_fields")
        # `_field_types` was removed in Python >=3.9.
        # and hasattr(cls, "_field_types")
        and hasattr(cls, "_field_defaults")
    )


def type_from_typevar_constraints(typ: TypeOrCallable) -> TypeOrCallable:
    """Try to concretize a type from a TypeVar's bounds or constraints. Identity if
    unsuccessful."""
    if isinstance(typ, TypeVar):
        if typ.__bound__ is not None:
            # Try to infer type from TypeVar bound.
            return typ.__bound__
        elif len(typ.__constraints__) > 0:
            # Try to infer type from TypeVar constraints.
            return Union.__getitem__(typ.__constraints__)  # type: ignore
    return typ


TypeOrCallableOrNone = TypeVar("TypeOrCallableOrNone", Callable, TypeForm[Any], None)


def unwrap_newtype(
    typ: TypeOrCallableOrNone,
) -> Tuple[TypeOrCallableOrNone, Optional[str]]:
    # We'll unwrap NewType annotations here; this is needed before issubclass
    # checks!
    #
    # `isinstance(x, NewType)` doesn't work because NewType isn't a class until
    # Python 3.10, so we instead do a duck typing-style check.
    return_name = None
    while hasattr(typ, "__name__") and hasattr(typ, "__supertype__"):
        if return_name is None:
            return_name = getattr(typ, "__name__")
        typ = getattr(typ, "__supertype__")

    return typ, return_name


@_unsafe_cache.unsafe_cache(maxsize=1024)
def unwrap_newtype_and_narrow_subtypes(
    typ: TypeOrCallable,
    default_instance: Any,
) -> TypeOrCallable:
    """Type narrowing: if we annotate as Animal but specify a default instance of Cat,
    we should parse as Cat.

    This should generally only be applied to fields used as nested structures, not
    individual arguments/fields. (if a field is annotated as Union[int, str], and a
    string default is passed in, we don't want to narrow the type to always be
    strings!)"""

    typ, unused_name = unwrap_newtype(typ)
    del unused_name

    try:
        potential_subclass = type(default_instance)

        if potential_subclass is type:
            # Don't narrow to `type`. This happens when the default instance is a class;
            # it doesn't really make sense to parse this case.
            return typ

        superclass = unwrap_annotated(typ)[0]

        # For Python 3.10.
        if get_origin(superclass) is Union:
            return typ

        if superclass is Any or issubclass(potential_subclass, superclass):  # type: ignore
            if get_origin(typ) is Annotated:
                return Annotated.__class_getitem__(  # type: ignore
                    (potential_subclass,) + get_args(typ)[1:]
                )
            typ = cast(TypeOrCallable, potential_subclass)
    except TypeError:
        # TODO: document where this TypeError can be raised, and reduce the amount of
        # code in it.
        pass

    return typ


def narrow_collection_types(
    typ: TypeOrCallable, default_instance: Any
) -> TypeOrCallable:
    """TypeForm narrowing for containers. Infers types of container contents."""
    if hasattr(default_instance, "__len__") and len(default_instance) == 0:
        return typ
    if typ is list and isinstance(default_instance, list):
        typ = List.__getitem__(Union.__getitem__(tuple(map(type, default_instance))))  # type: ignore
    elif typ is set and isinstance(default_instance, set):
        typ = Set.__getitem__(Union.__getitem__(tuple(map(type, default_instance))))  # type: ignore
    elif typ is tuple and isinstance(default_instance, tuple):
        typ = Tuple.__getitem__(tuple(map(type, default_instance)))  # type: ignore
    return typ


MetadataType = TypeVar("MetadataType")


def unwrap_annotated(
    typ: TypeOrCallable,
    search_type: TypeForm[MetadataType] = Any,  # type: ignore
) -> Tuple[TypeOrCallable, Tuple[MetadataType, ...]]:
    """Helper for parsing typing.Annotated types.

    Examples:
    - int, int => (int, ())
    - Annotated[int, 1], int => (int, (1,))
    - Annotated[int, "1"], int => (int, ())
    """
    # Check for __tyro_markers__ from @configure.
    targets = tuple(
        x
        for x in getattr(typ, "__tyro_markers__", tuple())
        if search_type is Any or isinstance(x, search_type)
    )
    assert isinstance(targets, tuple)
    if not hasattr(typ, "__metadata__"):
        return typ, targets  # type: ignore

    args = get_args(typ)
    assert len(args) >= 2

    # Look through metadata for desired metadata type.
    targets += tuple(
        x
        for x in targets + args[1:]
        if search_type is Any or isinstance(x, search_type)
    )

    # Check for __tyro_markers__ in unwrapped type.
    targets += tuple(
        x
        for x in getattr(args[0], "__tyro_markers__", tuple())
        if search_type is Any or isinstance(x, search_type)
    )
    return args[0], targets  # type: ignore


def apply_type_from_typevar(
    typ: TypeOrCallable, type_from_typevar: Dict[TypeVar, TypeForm[Any]]
) -> TypeOrCallable:
    if typ in type_from_typevar:
        return type_from_typevar[typ]  # type: ignore

    origin = get_origin(typ)
    args = get_args(typ)
    if len(args) > 0:
        if origin is Annotated:
            args = args[:1]
        if origin is collections.abc.Callable:
            assert isinstance(args[0], list)
            args = tuple(args[0]) + args[1:]

        # Convert Python 3.9 and 3.10 types to their typing library equivalents, which
        # support `.copy_with()`.
        if sys.version_info[:2] >= (3, 9):
            shim_table = {
                # PEP 585. Requires Python 3.9.
                tuple: Tuple,
                list: List,
                dict: Dict,
                set: Set,
                frozenset: FrozenSet,
            }
            if hasattr(types, "UnionType"):  # type: ignore
                # PEP 604. Requires Python 3.10.
                shim_table[types.UnionType] = Union  # type: ignore

            for new, old in shim_table.items():
                if isinstance(typ, new) or origin is new:  # type: ignore
                    typ = old.__getitem__(args)  # type: ignore

        return typ.copy_with(  # type: ignore
            tuple(apply_type_from_typevar(x, type_from_typevar) for x in args)
        )

    return typ


@_unsafe_cache.unsafe_cache(maxsize=1024)
def narrow_union_type(typ: TypeOrCallable, default_instance: Any) -> TypeOrCallable:
    """Narrow union types.

    This is a shim for failing more gracefully when we we're given a Union type that
    doesn't match the default value.

    In this case, we raise a warning, then add the type of the default value to the
    union. Loosely motivated by: https://github.com/brentyi/tyro/issues/20
    """
    if get_origin(typ) is not Union:
        return typ

    options = get_args(typ)
    options_unwrapped = [unwrap_origin_strip_extras(o) for o in options]

    # (A)
    try:
        if default_instance not in _fields.MISSING_SINGLETONS and not any(
            isinstance(default_instance, o) for o in options_unwrapped
        ):
            warnings.warn(
                f"{type(default_instance)} does not match any type in Union:"
                f" {options_unwrapped}"
            )
            return Union.__getitem__(options + (type(default_instance),))  # type: ignore
    except TypeError:
        pass

    return typ


def get_type_hints_with_nicer_errors(
    obj: Callable[..., Any], include_extras: bool = False
) -> Dict[str, Any]:
    try:
        return get_type_hints(obj, include_extras=include_extras)
    except TypeError as e:  # pragma: no cover
        common_message = f"\n-----\n\nError calling typing.get_type_hints() on {obj}! "
        message = e.args[0]
        if message.startswith(
            "unsupported operand type(s) for |"
        ) and sys.version_info < (3, 10):
            # PEP 604. Requires Python 3.10.
            raise TypeError(
                e.args[0]
                + common_message
                + "You may be using a Union in the form of `X | Y`; to support Python versions that lower than 3.10, you need to use `typing.Union[X, Y]` instead of `X | Y` and `typing.Optional[X]` instead of `X | None`.",
                *e.args[1:],
            )
        elif message == "'type' object is not subscriptable" and (
            sys.version_info < (3, 9)
        ):
            # PEP 585. Requires Python 3.9.
            raise TypeError(
                e.args[0]
                + common_message
                + "You may be using a standard collection as a generic, like `list[T]` or `tuple[T1, T2]`. To support Python versions lower than 3.9, you need to using `typing.List[T]` and `typing.Tuple[T1, T2]`.",
                *e.args[1:],
            )
        raise e



class Singleton:
    # Singleton pattern.
    # https://www.python.org/download/releases/2.2/descrintro/#__new__
    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it.init(*args, **kwds)
        return it

    def init(self, *args, **kwds):
        pass



import contextlib
import dataclasses
import io
from typing import Union

import pytest

import tyro


# https://github.com/brentyi/tyro/issues/9
@dataclasses.dataclass(frozen=True)
class Subtype:
    data: int = 1


@dataclasses.dataclass(frozen=True)
class TypeA:
    subtype: Subtype = Subtype(1)


@dataclasses.dataclass(frozen=True)
class TypeB:
    subtype: Subtype = Subtype(2)


@dataclasses.dataclass(frozen=True)
class Wrapper:
    supertype: Union[TypeA, TypeB] = TypeA()


def test_bash():
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(Wrapper, args=["--tyro-print-completion", "bash"])
    assert "# AUTOMATICALLY GENERATED by `shtab`" in target.getvalue()


def test_zsh():
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(Wrapper, args=["--tyro-print-completion", "zsh"])
    assert "# AUTOMATICALLY GENERATED by `shtab`" in target.getvalue()



import dataclasses

import tyro


def test_flag_default_false() -> None:
    """Test for argparse.BooleanOptionalAction-style usage."""

    @dataclasses.dataclass
    class A:
        x: bool

    assert tyro.cli(
        A,
        args=["--x"],
        default=A(False),
    ) == A(True)

    assert tyro.cli(
        A,
        args=["--no-x"],
        default=A(False),
    ) == A(False)

    assert tyro.cli(
        A,
        args=[],
        default=A(False),
    ) == A(False)

    assert tyro.cli(
        tyro.conf.FlagConversionOff[A],
        args=["--x", "True"],
        default=A(False),
    ) == A(True)


def test_flag_default_true() -> None:
    """Test for argparse.BooleanOptionalAction-style usage."""

    @dataclasses.dataclass
    class A:
        x: bool

    assert tyro.cli(
        A,
        args=["--x"],
        default=A(True),
    ) == A(True)

    assert tyro.cli(
        A,
        args=["--no-x"],
        default=A(True),
    ) == A(False)

    assert tyro.cli(
        A,
        args=[],
        default=A(True),
    ) == A(True)

    assert tyro.cli(
        tyro.conf.FlagConversionOff[A],
        args=["--x", "True"],
        default=A(False),
    ) == A(True)



import dataclasses
from typing import Any, Callable, Optional, Sequence, Tuple, Type, Union, overload

from .._fields import MISSING_NONPROP


@dataclasses.dataclass(frozen=True)
class _SubcommandConfiguration:
    name: Optional[str]
    default: Any
    description: Optional[str]
    prefix_name: bool
    constructor_factory: Optional[Callable[[], Union[Type, Callable]]]

    def __hash__(self) -> int:
        return object.__hash__(self)


@overload
def subcommand(
    name: Optional[str] = None,
    *,
    default: Any = MISSING_NONPROP,
    description: Optional[str] = None,
    prefix_name: bool = True,
    constructor: None = None,
    constructor_factory: Optional[Callable[[], Union[Type, Callable]]] = None,
) -> Any:
    ...


@overload
def subcommand(
    name: Optional[str] = None,
    *,
    default: Any = MISSING_NONPROP,
    description: Optional[str] = None,
    prefix_name: bool = True,
    constructor: Optional[Union[Type, Callable]] = None,
    constructor_factory: None = None,
) -> Any:
    ...


def subcommand(
    name: Optional[str] = None,
    *,
    default: Any = MISSING_NONPROP,
    description: Optional[str] = None,
    prefix_name: bool = True,
    constructor: Optional[Union[Type, Callable]] = None,
    constructor_factory: Optional[Callable[[], Union[Type, Callable]]] = None,
) -> Any:
    """Returns a metadata object for configuring subcommands with `typing.Annotated`.
    Useful for aesthetics.

    Consider the standard approach for creating subcommands:

    ```python
    tyro.cli(
        Union[NestedTypeA, NestedTypeB]
    )
    ```

    This will create two subcommands: `nested-type-a` and `nested-type-b`.

    Annotating each type with `tyro.conf.subcommand()` allows us to override for
    each subcommand the (a) name, (b) defaults, (c) helptext, and (d) whether to prefix
    the name or not.

    ```python
    tyro.cli(
        Union[
            Annotated[
                NestedTypeA, subcommand("a", ...)
            ],
            Annotated[
                NestedTypeB, subcommand("b", ...)
            ],
        ]
    )
    ```

    Arguments:
        name: The name of the subcommand in the CLI.
        default: A default value for the subcommand, for struct-like types. (eg
             dataclasses)
        description: Description of this option to use in the helptext. Defaults to
            docstring.
        prefix_name: Whether to prefix the name of the subcommand based on where it
            is in a nested structure.
        constructor: A constructor type or function. This should either be (a) a subtype
            of an argument's annotated type, or (b) a function with type-annotated
            inputs that returns an instance of the annotated type. This will be used in
            place of the argument's type for parsing arguments. No validation is done.
        constructor_factory: A function that returns a constructor type or function.
            Useful when the constructor isn't immediately available.
    """
    assert not (
        constructor is not None and constructor_factory is not None
    ), "`constructor` and `constructor_factory` cannot both be set."
    return _SubcommandConfiguration(
        name,
        default,
        description,
        prefix_name,
        constructor_factory=constructor_factory
        if constructor is None
        else lambda: constructor,
    )


@dataclasses.dataclass(frozen=True)
class _ArgConfiguration:
    # These are all optional by default in order to support multiple tyro.conf.arg()
    # annotations. A None value means "don't overwrite the current value".
    name: Optional[str]
    metavar: Optional[str]
    help: Optional[str]
    aliases: Optional[Tuple[str, ...]]
    prefix_name: Optional[bool]
    constructor_factory: Optional[Callable[[], Union[Type, Callable]]]


@overload
def arg(
    *,
    name: Optional[str] = None,
    metavar: Optional[str] = None,
    help: Optional[str] = None,
    aliases: Optional[Sequence[str]] = None,
    prefix_name: Optional[bool] = None,
    constructor: None = None,
    constructor_factory: Optional[Callable[[], Union[Type, Callable]]] = None,
) -> Any:
    ...


@overload
def arg(
    *,
    name: Optional[str] = None,
    metavar: Optional[str] = None,
    help: Optional[str] = None,
    aliases: Optional[Sequence[str]] = None,
    prefix_name: Optional[bool] = None,
    constructor: Optional[Union[Type, Callable]] = None,
    constructor_factory: None = None,
) -> Any:
    ...


def arg(
    *,
    name: Optional[str] = None,
    metavar: Optional[str] = None,
    help: Optional[str] = None,
    aliases: Optional[Sequence[str]] = None,
    prefix_name: Optional[bool] = None,
    constructor: Optional[Union[Type, Callable]] = None,
    constructor_factory: Optional[Callable[[], Union[Type, Callable]]] = None,
) -> Any:
    """Returns a metadata object for fine-grained argument configuration with
    `typing.Annotated`. Should typically not be required.
    ```python
    x: Annotated[int, tyro.conf.arg(...)]
    ```

    Arguments:
        name: A new name for the argument in the CLI.
        metavar: Argument name in usage messages. The type is used by default.
        help: Helptext for this argument. The docstring is used by default.
        aliases: Aliases for this argument. All strings in the sequence should start
            with a hyphen (-). Aliases will _not_ currently be prefixed in a nested
            structure, and are not supported for positional arguments.
        prefix_name: Whether or not to prefix the name of the argument based on where
            it is in a nested structure. Arguments are prefixed by default.
        constructor: A constructor type or function. This should either be (a) a subtype
            of an argument's annotated type, or (b) a function with type-annotated
            inputs that returns an instance of the annotated type. This will be used in
            place of the argument's type for parsing arguments. No validation is done.
        constructor_factory: A function that returns a constructor type or function.
            Useful when the constructor isn't immediately available.

    Returns:
        Object to attach via `typing.Annotated[]`.
    """
    assert not (
        constructor is not None and constructor_factory is not None
    ), "`constructor` and `constructor_factory` cannot both be set."

    if aliases is not None:
        for alias in aliases:
            assert alias.startswith("-"), "Argument alias needs to start with a hyphen!"

    return _ArgConfiguration(
        name=name,
        metavar=metavar,
        help=help,
        aliases=tuple(aliases) if aliases is not None else None,
        prefix_name=prefix_name,
        constructor_factory=constructor_factory
        if constructor is None
        else lambda: constructor,
    )



"""Core public API."""
import argparse
import dataclasses
import pathlib
import sys
import warnings
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import shtab
from typing_extensions import Literal

from . import (
    _argparse_formatter,
    _arguments,
    _calling,
    _fields,
    _parsers,
    _strings,
    _unsafe_cache,
    conf,
)
from ._typing import TypeForm

OutT = TypeVar("OutT")


# Note that the overload here is necessary for pyright and pylance due to special-casing
# related to using typing.Type[] as a temporary replacement for typing.TypeForm[].
#
# https://github.com/microsoft/pyright/issues/4298


@overload
def cli(
    f: TypeForm[OutT],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    default: Optional[OutT] = None,
    return_unknown_args: Literal[False] = False,
    use_underscores: bool = False,
) -> OutT:
    ...


@overload
def cli(
    f: TypeForm[OutT],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    default: Optional[OutT] = None,
    return_unknown_args: Literal[True],
    use_underscores: bool = False,
) -> Tuple[OutT, List[str]]:
    ...


@overload
def cli(
    f: Callable[..., OutT],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    # Note that passing a default makes sense for things like dataclasses, but are not
    # supported for general callables. These can, however, be specified in the signature
    # of the callable itself.
    default: None = None,
    return_unknown_args: Literal[False] = False,
    use_underscores: bool = False,
) -> OutT:
    ...


@overload
def cli(
    f: Callable[..., OutT],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    # Note that passing a default makes sense for things like dataclasses, but are not
    # supported for general callables. These can, however, be specified in the signature
    # of the callable itself.
    default: None = None,
    return_unknown_args: Literal[True],
    use_underscores: bool = False,
) -> Tuple[OutT, List[str]]:
    ...


def cli(
    f: Union[TypeForm[OutT], Callable[..., OutT]],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    default: Optional[OutT] = None,
    return_unknown_args: bool = False,
    use_underscores: bool = False,
    **deprecated_kwargs,
) -> Union[OutT, Tuple[OutT, List[str]]]:
    """Call or instantiate `f`, with inputs populated from an automatically generated
    CLI interface.

    `f` should have type-annotated inputs, and can be a function or type. Note that if
    `f` is a type, `tyro.cli()` returns an instance.

    The parser is generated by populating helptext from docstrings and types from
    annotations; a broad range of core type annotations are supported.
    - Types natively accepted by `argparse`: str, int, float, pathlib.Path, etc.
    - Default values for optional parameters.
    - Booleans, which are automatically converted to flags when provided a default
      value.
    - Enums (via `enum.Enum`).
    - Various annotations from the standard typing library. Some examples:
      - `typing.ClassVar[T]`.
      - `typing.Optional[T]`.
      - `typing.Literal[T]`.
      - `typing.Sequence[T]`.
      - `typing.List[T]`.
      - `typing.Dict[K, V]`.
      - `typing.Tuple`, such as `typing.Tuple[T1, T2, T3]` or
        `typing.Tuple[T, ...]`.
      - `typing.Set[T]`.
      - `typing.Final[T]` and `typing.Annotated[T]`.
      - `typing.Union[T1, T2]`.
      - Various nested combinations of the above: `Optional[Literal[T]]`,
        `Final[Optional[Sequence[T]]]`, etc.
    - Hierarchical structures via nested dataclasses, TypedDict, NamedTuple,
      classes.
      - Simple nesting.
      - Unions over nested structures (subparsers).
      - Optional unions over nested structures (optional subparsers).
    - Generics (including nested generics).

    Completion script generation for interactive shells is also provided. To write a
    script that can be used for tab completion, pass in:
        `--tyro-write-completion {bash/zsh/tcsh} {path to script to write}`.

    Args:
        f: Function or type.
        prog: The name of the program printed in helptext. Mirrors argument from
            `argparse.ArgumentParser()`.
        description: Description text for the parser, displayed when the --help flag is
            passed in. If not specified, `f`'s docstring is used. Mirrors argument from
            `argparse.ArgumentParser()`.
        args: If set, parse arguments from a sequence of strings instead of the
            commandline. Mirrors argument from `argparse.ArgumentParser.parse_args()`.
        default: An instance of `OutT` to use for default values; supported if `f` is a
            type like a dataclass or dictionary, but not if `f` is a general callable like
            a function or standard class. Helpful for merging CLI arguments with values
            loaded from elsewhere. (for example, a config object loaded from a yaml file)
        return_unknown_args: If True, return a tuple of the output of `f` and a list of
            unknown arguments. Mirrors the unknown arguments returned from
            `argparse.ArgumentParser.parse_known_args()`.
        use_underscores: If True, use underscores as a word delimeter instead of hyphens.
            This primarily impacts helptext; underscores and hyphens are treated equivalently
            when parsing happens. We default helptext to hyphens to follow the GNU style guide.
            https://www.gnu.org/software/libc/manual/html_node/Argument-Syntax.html

    Returns:
        The output of `f(...)` or an instance `f`. If `f` is a class, the two are
        equivalent. If `return_unknown_args` is True, returns a tuple of the output of
        `f(...)` and a list of unknown arguments.
    """

    # Make sure we start on a clean slate. Some tests may fail without this due to
    # memory address conflicts.
    _unsafe_cache.clear_cache()

    with _strings.delimeter_context("_" if use_underscores else "-"):
        output = _cli_impl(
            f,
            prog=prog,
            description=description,
            args=args,
            default=default,
            return_parser=False,
            return_unknown_args=return_unknown_args,
            use_underscores=use_underscores,
            **deprecated_kwargs,
        )

    # Prevent unnecessary memory usage.
    _unsafe_cache.clear_cache()

    if return_unknown_args:
        return cast(Tuple[OutT, List[str]], output)
    else:
        return cast(OutT, output)


@overload
def get_parser(
    f: TypeForm[OutT],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    default: Optional[OutT] = None,
    use_underscores: bool = False,
) -> argparse.ArgumentParser:
    ...


@overload
def get_parser(
    f: Callable[..., OutT],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    default: Optional[OutT] = None,
    use_underscores: bool = False,
) -> argparse.ArgumentParser:
    ...


def get_parser(
    f: Union[TypeForm[OutT], Callable[..., OutT]],
    *,
    # Note that we have no `args` argument, since this is only used when
    # parser.parse_args() is called.
    prog: Optional[str] = None,
    description: Optional[str] = None,
    default: Optional[OutT] = None,
    use_underscores: bool = False,
) -> argparse.ArgumentParser:
    """Get the `argparse.ArgumentParser` object generated under-the-hood by
    `tyro.cli()`. Useful for tools like `sphinx-argparse`, `argcomplete`, etc.

    For tab completion, we recommend using `tyro.cli()`'s built-in `--tyro-write-completion`
    flag."""
    with _strings.delimeter_context("_" if use_underscores else "-"):
        return cast(
            argparse.ArgumentParser,
            _cli_impl(
                f,
                prog=prog,
                description=description,
                args=None,
                default=default,
                return_parser=True,
                return_unknown_args=False,
                use_underscores=use_underscores,
            ),
        )


def _cli_impl(
    f: Union[TypeForm[OutT], Callable[..., OutT]],
    *,
    prog: Optional[str] = None,
    description: Optional[str],
    args: Optional[Sequence[str]],
    default: Optional[OutT],
    return_parser: bool,
    return_unknown_args: bool,
    **deprecated_kwargs,
) -> Union[
    OutT,
    argparse.ArgumentParser,
    Tuple[OutT, List[str]],
]:
    """Helper for stitching the `tyro` pipeline together."""
    if "default_instance" in deprecated_kwargs:
        warnings.warn(
            "`default_instance=` is deprecated! use `default=` instead.", stacklevel=2
        )
        default = deprecated_kwargs["default_instance"]
    if deprecated_kwargs.get("avoid_subparsers", False):
        f = conf.AvoidSubcommands[f]  # type: ignore
        warnings.warn(
            "`avoid_subparsers=` is deprecated! use `tyro.conf.AvoidSubparsers[]`"
            " instead.",
            stacklevel=2,
        )

    # Internally, we distinguish between two concepts:
    # - "default", which is used for individual arguments.
    # - "default_instance", which is used for _fields_ (which may be broken down into
    #   one or many arguments, depending on various factors).
    #
    # This could be revisited.
    default_instance_internal: Union[_fields.NonpropagatingMissingType, OutT] = (
        _fields.MISSING_NONPROP if default is None else default
    )

    # We wrap our type with a dummy dataclass if it can't be treated as a nested type.
    # For example: passing in f=int will result in a dataclass with a single field
    # typed as int.
    if not _fields.is_nested_type(cast(type, f), default_instance_internal):
        dummy_field = cast(
            dataclasses.Field,
            dataclasses.field(),
        )
        f = dataclasses.make_dataclass(
            cls_name="dummy",
            fields=[(_strings.dummy_field_name, cast(type, f), dummy_field)],
            frozen=True,
        )
        default_instance_internal = f(default_instance_internal)  # type: ignore
        dummy_wrapped = True
    else:
        dummy_wrapped = False

    # Read and fix arguments. If the user passes in --field_name instead of
    # --field-name, correct for them.
    args = list(sys.argv[1:]) if args is None else list(args)

    # Fix arguments. This will modify all option-style arguments replacing
    # underscores with hyphens, or vice versa if use_underscores=True.
    # If two options are ambiguous, e.g., --a_b and --a-b, raise a runtime error.
    modified_args: Dict[str, str] = {}
    for index, arg in enumerate(args):
        if not arg.startswith("--"):
            continue

        delimeter = _strings.get_delimeter()
        to_swap_delimeter = "-" if delimeter == "_" else "_"

        if "=" in arg:
            arg, _, val = arg.partition("=")
            fixed = "--" + arg[2:].replace(to_swap_delimeter, delimeter) + "=" + val
        else:
            fixed = "--" + arg[2:].replace(to_swap_delimeter, delimeter)
        if (
            return_unknown_args
            and fixed in modified_args
            and modified_args[fixed] != arg
        ):
            raise RuntimeError(
                "Ambiguous arguments: " + modified_args[fixed] + " and " + arg
            )
        modified_args[fixed] = arg
        args[index] = fixed

    # If we pass in the --tyro-print-completion or --tyro-write-completion flags: turn
    # formatting tags, and get the shell we want to generate a completion script for
    # (bash/zsh/tcsh).
    #
    # shtab also offers an add_argument_to() functions that fulfills a similar goal, but
    # manual parsing of argv is convenient for turning off formatting.
    #
    # Note: --tyro-print-completion is deprecated! --tyro-write-completion is less prone
    # to errors from accidental logging, print statements, etc.
    print_completion = len(args) >= 2 and args[0] == "--tyro-print-completion"
    write_completion = len(args) >= 3 and args[0] == "--tyro-write-completion"

    # Note: setting USE_RICH must happen before the parser specification is generated.
    # TODO: revisit this. Ideally we should be able to eliminate the global state
    # changes.
    completion_shell = None
    completion_target_path = None
    if print_completion or write_completion:
        completion_shell = args[1]
    if write_completion:
        completion_target_path = pathlib.Path(args[2])
    if print_completion or write_completion or return_parser:
        _arguments.USE_RICH = False
    else:
        _arguments.USE_RICH = True

    # Map a callable to the relevant CLI arguments + subparsers.
    parser_spec = _parsers.ParserSpecification.from_callable_or_type(
        f,
        description=description,
        parent_classes=set(),  # Used for recursive calls.
        default_instance=default_instance_internal,  # Overrides for default values.
        intern_prefix="",  # Used for recursive calls.
        extern_prefix="",  # Used for recursive calls.
        subcommand_prefix="",  # Used for recursive calls.
    )

    # Generate parser!
    with _argparse_formatter.ansi_context():
        parser = _argparse_formatter.TyroArgumentParser(
            prog=prog,
            formatter_class=_argparse_formatter.TyroArgparseHelpFormatter,
            allow_abbrev=False,
        )
        parser._parser_specification = parser_spec
        parser._parsing_known_args = return_unknown_args
        parser._args = args
        parser_spec.apply(parser)

        # Print help message when no arguments are passed in. (but arguments are
        # expected)
        # if len(args) == 0 and parser_spec.has_required_args:
        #     args = ["--help"]

        if return_parser:
            _arguments.USE_RICH = True
            return parser

        if print_completion or write_completion:
            _arguments.USE_RICH = True
            assert completion_shell in (
                "bash",
                "zsh",
                "tcsh",
            ), (
                "Shell should be one `bash`, `zsh`, or `tcsh`, but got"
                f" {completion_shell}"
            )

            if write_completion:
                assert completion_target_path is not None
                completion_target_path.write_text(
                    shtab.complete(
                        parser=parser,
                        shell=completion_shell,
                        root_prefix=f"tyro_{parser.prog}",
                    )
                )
            else:
                print(
                    shtab.complete(
                        parser=parser,
                        shell=completion_shell,
                        root_prefix=f"tyro_{parser.prog}",
                    )
                )
            sys.exit()

        if return_unknown_args:
            namespace, unknown_args = parser.parse_known_args(args=args)
        else:
            unknown_args = None
            namespace = parser.parse_args(args=args)
        value_from_prefixed_field_name = vars(namespace)

    if dummy_wrapped:
        value_from_prefixed_field_name = {
            k.replace(_strings.dummy_field_name, ""): v
            for k, v in value_from_prefixed_field_name.items()
        }

    try:
        # Attempt to call `f` using whatever was passed in.
        out, consumed_keywords = _calling.call_from_args(
            f,
            parser_spec,
            default_instance_internal,
            value_from_prefixed_field_name,
            field_name_prefix="",
        )
    except _calling.InstantiationError as e:
        assert isinstance(e, _calling.InstantiationError)

        # Emulate argparse's error behavior when invalid arguments are passed in.
        from rich.console import Console, Group, RenderableType
        from rich.padding import Padding
        from rich.panel import Panel
        from rich.rule import Rule
        from rich.style import Style

        from ._argparse_formatter import THEME

        console = Console(theme=THEME.as_rich_theme())
        console.print(
            Panel(
                Group(
                    "[bright_red][bold]Error parsing"
                    f" {e.arg.lowered.name_or_flag if isinstance(e.arg, _arguments.ArgumentDefinition) else e.arg}[/bold]:[/bright_red] {e.message}",
                    *cast(  # Cast to appease mypy...
                        List[RenderableType],
                        (
                            []
                            if not isinstance(e.arg, _arguments.ArgumentDefinition)
                            or e.arg.lowered.help is None
                            else [
                                Rule(style=Style(color="red")),
                                "Argument helptext:",
                                Padding(
                                    Group(
                                        f"{e.arg.lowered.name_or_flag} [bold]{e.arg.lowered.metavar}[/bold]",
                                        e.arg.lowered.help,
                                    ),
                                    pad=(0, 0, 0, 4),
                                ),
                                Rule(style=Style(color="red")),
                                f"For full helptext, see [bold]{parser.prog} --help[/bold]",
                            ]
                        ),
                    ),
                ),
                title="[bold]Value error[/bold]",
                title_align="left",
                border_style=Style(color="red"),
            )
        )
        sys.exit(2)

    assert len(value_from_prefixed_field_name.keys() - consumed_keywords) == 0, (
        f"Parsed {value_from_prefixed_field_name.keys()}, but only consumed"
        f" {consumed_keywords}"
    )

    if dummy_wrapped:
        out = getattr(out, _strings.dummy_field_name)

    if return_unknown_args:
        assert unknown_args is not None, "Should have parsed with `parse_known_args()`"
        # If we're parsed unknown args, we should return the original args, not
        # the fixed ones.
        unknown_args = [modified_args.get(arg, arg) for arg in unknown_args]
        return out, unknown_args
    else:
        assert unknown_args is None, "Should have parsed with `parse_args()`"
        return out



"""Helper for using type annotations to recursively generate instantiator functions,
which map sequences of strings to the annotated type.

Some examples of type annotations and the desired instantiators:
```
    int

        lambda strings: int(str[0])

    List[int]

        lambda strings: list(
            [int(x) for x in strings]
        )

    List[Color], where Color is an enum

        lambda strings: list(
            [Color[x] for x in strings]
        )

    Tuple[int, float]

        lambda strings: tuple(
            typ(x)
            for typ, x in zip(
                (int, float),
                strings,
            )
        )
```
"""
import collections.abc
import dataclasses
import enum
import inspect
import os
import pathlib
from collections import deque
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Hashable,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from typing_extensions import Annotated, Final, Literal, get_args, get_origin

from . import _resolver

# There are cases where typing.Literal doesn't match typing_extensions.Literal:
# https://github.com/python/typing_extensions/pull/148
try:
    from typing import Literal as LiteralAlternate
except ImportError:
    LiteralAlternate = Literal  # type: ignore

from . import _strings
from ._typing import TypeForm
from .conf import _markers

_StandardInstantiator = Callable[[List[str]], Any]
_AppendNargsInstantiator = Callable[[List[List[str]]], Any]
# Special case: the only time that argparse doesn't give us a string is when the
# argument action is set to `store_true` or `store_false`. In this case, we get
# a bool directly, and the field action can be a no-op.
_FlagInstantiator = Callable[[bool], bool]

Instantiator = Union[_StandardInstantiator, _AppendNargsInstantiator, _FlagInstantiator]

NoneType = type(None)


@dataclasses.dataclass
class InstantiatorMetadata:
    # Unlike in vanilla argparse, we never set nargs to None. To make things simpler, we
    # instead use nargs=1.
    nargs: Optional[Union[int, Literal["*"]]]
    # Unlike in vanilla argparse, our metavar is always a string. We handle
    # sequences, multiple arguments, etc, manually.
    metavar: str
    choices: Optional[Set[str]]
    action: Optional[Literal["append"]]

    def check_choices(self, strings: List[str]) -> None:
        if self.choices is not None and any(s not in self.choices for s in strings):
            raise ValueError(f"invalid choice: {strings} (choose from {self.choices}))")


class UnsupportedTypeAnnotationError(Exception):
    """Exception raised when an unsupported type annotation is detected."""


_builtin_set = set(
    filter(
        lambda x: isinstance(x, Hashable),  # type: ignore
        __builtins__.values(),  # type: ignore
    )
)


def is_type_string_converter(typ: Union[Callable, TypeForm[Any]]) -> bool:
    """Check if type is a string converter, i.e., (arg: Union[str, Any]) -> T."""
    param_count = 0
    has_var_positional = False
    try:
        signature = inspect.signature(typ)
    except ValueError:
        # pybind objects might not have a parsable signature. We try to be tolerant in this case.
        # One day this should be fixed with `__text_signature__`.
        return True

    type_annotations = _resolver.get_type_hints_with_nicer_errors(typ)
    # Some checks we can do if the signature is available!
    for i, param in enumerate(signature.parameters.values()):
        annotation = type_annotations.get(param.name, param.annotation)

        # Hack: apply_type_from_typevar applies shims, like UnionType => Union
        # conversion.
        annotation = _resolver.apply_type_from_typevar(annotation, {})
        if i == 0 and not (
            (get_origin(annotation) is Union and str in get_args(annotation))
            or annotation in (str, inspect.Parameter.empty)
        ):
            return False
        if param.kind is inspect.Parameter.VAR_POSITIONAL:
            has_var_positional = True
        elif param.default is inspect.Parameter.empty and param.kind is not (
            inspect.Parameter.VAR_KEYWORD
        ):
            param_count += 1

    # Raise an error if parameters look wrong.
    if not (param_count == 1 or (param_count == 0 and has_var_positional)):
        return False
    return True


def instantiator_from_type(
    typ: Union[TypeForm[Any], Callable],
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[Instantiator, InstantiatorMetadata]:
    """Recursive helper for parsing type annotations.

    Returns two things:
    - An instantiator function, for instantiating the type from a string or list of
      strings. The latter applies when argparse's `nargs` parameter is set.
    - A metadata structure, which specifies parameters for argparse.
    """

    # Handle Any.
    if typ is Any:
        raise UnsupportedTypeAnnotationError("`Any` is not a parsable type.")

    # Handle NoneType.
    if typ is NoneType:

        def instantiator(strings: List[str]) -> None:
            # Note that other inputs should be caught by `choices` before the
            # instantiator runs.
            assert strings == ["None"]

        return instantiator, InstantiatorMetadata(
            nargs=1,
            metavar="{None}",
            choices={
                "None",
            },
            action=None,
        )

    # Instantiate os.PathLike annotations using pathlib.Path.
    # Ideally this should be implemented in a more general way, eg using
    # https://github.com/brentyi/tyro/pull/30
    if typ is os.PathLike:
        typ = pathlib.Path

    # Address container types. If a matching container is found, this will recursively
    # call instantiator_from_type().
    container_out = _instantiator_from_container_type(
        cast(TypeForm[Any], typ), type_from_typevar, markers
    )
    if container_out is not None:
        return container_out

    # Unwrap NewType + set metavar based on NewType name.
    # `isinstance(x, NewType)` doesn't work because NewType isn't a class until
    # Python 3.10, so we instead do a duck typing-style check.
    metavar = getattr(typ, "__name__", "").upper()
    typ, maybe_newtype_name = _resolver.unwrap_newtype(typ)
    if maybe_newtype_name is not None:
        metavar = maybe_newtype_name.upper()

    # Validate that typ is a `(arg: str) -> T` type converter, as expected by argparse.
    if typ in _builtin_set:
        pass
    elif not callable(typ):
        raise UnsupportedTypeAnnotationError(
            f"Expected {typ} to be an `(arg: str) -> T` type converter, but is not"
            " callable."
        )
    elif not is_type_string_converter(typ):
        raise UnsupportedTypeAnnotationError(
            f"Expected {typ} to be an `(arg: str) -> T` type converter, but is not"
            " a valid type converter."
        )

    # Special case `choices` for some types, as implemented in `instance_from_string()`.
    auto_choices: Optional[Tuple[str, ...]] = None
    if typ is bool:
        auto_choices = ("True", "False")
    elif inspect.isclass(typ) and issubclass(typ, enum.Enum):
        auto_choices = tuple(x.name for x in typ)

    def instantiator_base_case(strings: List[str]) -> Any:
        """Given a type and and a string from the command-line, reconstruct an object. Not
        intended to deal with containers.

        This is intended to replace all calls to `type(string)`, which can cause unexpected
        behavior. As an example, note that the following argparse code will always print
        `True`, because `bool("True") == bool("False") == bool("0") == True`.
        ```
        import argparse

        parser = argparse.ArgumentParser()
        parser.add_argument("--flag", type=bool)

        print(parser.parse_args().flag)
        ```
        """
        assert len(get_args(typ)) == 0, f"TypeForm {typ} cannot be instantiated."
        (string,) = strings
        if typ is bool:
            return {"True": True, "False": False}[string]  # type: ignore
        elif isinstance(typ, type) and issubclass(typ, enum.Enum):
            return typ[string]  # type: ignore
        elif typ is bytes:
            return bytes(string, encoding="ascii")  # type: ignore
        else:
            return typ(string)  # type: ignore

    return instantiator_base_case, InstantiatorMetadata(
        nargs=1,
        metavar=(
            metavar
            if auto_choices is None
            else "{" + ",".join(map(str, auto_choices)) + "}"
        ),
        choices=None if auto_choices is None else set(auto_choices),
        action=None,
    )


@overload
def _instantiator_from_type_inner(
    typ: TypeForm,
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    allow_sequences: Literal["fixed_length"],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[Instantiator, InstantiatorMetadata]:
    ...


@overload
def _instantiator_from_type_inner(
    typ: TypeForm,
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    allow_sequences: Literal[False],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[_StandardInstantiator, InstantiatorMetadata]:
    ...


@overload
def _instantiator_from_type_inner(
    typ: TypeForm,
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    allow_sequences: Literal[True],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[Instantiator, InstantiatorMetadata]:
    ...


def _instantiator_from_type_inner(
    typ: TypeForm,
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    allow_sequences: Literal["fixed_length", True, False],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[Instantiator, InstantiatorMetadata]:
    """Thin wrapper over instantiator_from_type, with some extra asserts for catching
    errors."""
    out = instantiator_from_type(typ, type_from_typevar, markers)
    if out[1].nargs == "*":
        # We currently only use allow_sequences=False for options in Literal types,
        # which are evaluated using `type()`. It should not be possible to hit this
        # condition from polling a runtime type.
        assert allow_sequences
        if allow_sequences == "fixed_length" and not isinstance(out[1].nargs, int):
            raise UnsupportedTypeAnnotationError(
                f"{typ} is a variable-length sequence, which is ambiguous when nested."
                " For nesting variable-length sequences (example: List[List[int]]),"
                " `tyro.conf.UseAppendAction` can help resolve ambiguities."
            )
    return out


def _instantiator_from_container_type(
    typ: TypeForm[Any],
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    markers: FrozenSet[_markers.Marker],
) -> Optional[Tuple[Instantiator, InstantiatorMetadata]]:
    """Attempt to create an instantiator from a container type. Returns `None` is no
    container type is found."""

    # Default generic types to strings.
    if typ in (dict, Dict):
        typ = Dict[str, str]
    elif typ in (tuple, Tuple):
        typ = Tuple[str, ...]  # type: ignore
    elif typ in (list, List, collections.abc.Sequence, Sequence):
        typ = List[str]
    elif typ in (set, Set):
        typ = Set[str]

    type_origin = get_origin(typ)
    if type_origin is None:
        return None

    # Unwrap Annotated and Final types.
    if type_origin in (Annotated, Final):
        contained_type = get_args(typ)[0]
        return instantiator_from_type(contained_type, type_from_typevar, markers)

    for make, matched_origins in {
        _instantiator_from_sequence: (
            collections.abc.Sequence,
            frozenset,
            list,
            set,
            deque,
        ),
        _instantiator_from_tuple: (tuple,),
        _instantiator_from_dict: (dict, collections.abc.Mapping),
        _instantiator_from_union: (Union,),
        _instantiator_from_literal: (Literal, LiteralAlternate),
    }.items():
        if type_origin in matched_origins:
            return make(typ, type_from_typevar, markers)

    raise UnsupportedTypeAnnotationError(  # pragma: no cover
        f"Unsupported type {typ} with origin {type_origin}"
    )


def _instantiator_from_tuple(
    typ: TypeForm,
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[Instantiator, InstantiatorMetadata]:
    types = get_args(typ)
    typeset = set(types)  # Note that sets are unordered.
    typeset_no_ellipsis = typeset - {Ellipsis}  # type: ignore

    if typeset_no_ellipsis != typeset:
        # Ellipsis: variable argument counts. When an ellipsis is used, tuples must
        # contain only one type.
        assert len(typeset_no_ellipsis) == 1
        return _instantiator_from_sequence(typ, type_from_typevar, markers)

    else:
        instantiators: List[_StandardInstantiator] = []
        metas: List[InstantiatorMetadata] = []
        nargs = 0
        for t in types:
            a, b = _instantiator_from_type_inner(
                t, type_from_typevar, allow_sequences="fixed_length", markers=markers
            )
            instantiators.append(a)  # type: ignore
            metas.append(b)
            assert isinstance(b.nargs, int)
            nargs += b.nargs

        def fixed_length_tuple_instantiator(strings: List[str]) -> Any:
            assert len(strings) == nargs

            # Make tuple.
            out = []
            i = 0
            for make, meta in zip(instantiators, metas):
                assert isinstance(meta.nargs, int)
                meta.check_choices(strings[i : i + meta.nargs])
                out.append(make(strings[i : i + meta.nargs]))
                i += meta.nargs
            return tuple(out)

        return fixed_length_tuple_instantiator, InstantiatorMetadata(
            nargs=nargs,
            metavar=" ".join(m.metavar for m in metas),
            choices=None,
            action=None,
        )


def _join_union_metavars(metavars: Iterable[str]) -> str:
    """Metavar generation helper for unions. Could be revisited.

    Examples:
        None, INT => NONE|INT
        {0,1,2}, {3,4} => {0,1,2,3,4}
        {0,1,2}, {3,4}, STR => {0,1,2,3,4}|STR
        {None}, INT [INT ...] => {None}|{INT [INT ...]}
        STR, INT [INT ...] => STR|{INT [INT ...]}
        STR, INT INT => STR|{INT INT}

    The curly brackets are unfortunately overloaded but alternatives all interfere with
    argparse internals.
    """
    metavars = tuple(metavars)
    merged_metavars = [metavars[0]]
    for i in range(1, len(metavars)):
        prev = merged_metavars[-1]
        curr = metavars[i]
        if (
            prev.startswith("{")
            and prev.endswith("}")
            and curr.startswith("{")
            and curr.endswith("}")
        ):
            merged_metavars[-1] = prev[:-1] + "," + curr[1:]
        else:
            merged_metavars.append(curr)

    for i, m in enumerate(merged_metavars):
        if " " in m:
            merged_metavars[i] = "{" + m + "}"

    return "|".join(merged_metavars)


def _instantiator_from_union(
    typ: TypeForm,
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[Instantiator, InstantiatorMetadata]:
    options = list(get_args(typ))
    if NoneType in options:
        # Move `None` types to the beginning.
        # If we have `Optional[str]`, we want this to be parsed as
        # `Union[NoneType, str]`.
        options.remove(NoneType)
        options.insert(0, NoneType)

    # General unions, eg Union[int, bool]. We'll try to convert these from left to
    # right.
    instantiators = []
    metas = []
    nargs: Optional[Union[int, Literal["*"]]] = 1
    first = True
    for t in options:
        a, b = _instantiator_from_type_inner(
            t, type_from_typevar, allow_sequences=True, markers=markers
        )
        instantiators.append(a)
        metas.append(b)

        if t is not NoneType:
            # Enforce that `nargs` is the same for all child types, except for
            # NoneType.
            if first:
                nargs = b.nargs
                first = False
            elif nargs != b.nargs:
                # Just be as general as possible if we see inconsistencies.
                nargs = "*"

    metavar: str
    metavar = _join_union_metavars(map(lambda x: cast(str, x.metavar), metas))

    def union_instantiator(strings: List[str]) -> Any:
        metadata: InstantiatorMetadata
        errors = []
        for i, (instantiator, metadata) in enumerate(zip(instantiators, metas)):
            # Check choices.
            if metadata.choices is not None and any(
                x not in metadata.choices for x in strings
            ):
                errors.append(
                    f"{options[i]}: {strings} does not match choices {metadata.choices}"
                )
                continue

            # Try passing input into instantiator.
            if len(strings) == metadata.nargs or (metadata.nargs == "*"):
                try:
                    return instantiator(strings)  # type: ignore
                except ValueError as e:
                    # Failed, try next instantiator.
                    errors.append(f"{options[i]}: {e.args[0]}")
            else:
                errors.append(
                    f"{options[i]}: input length {len(strings)} did not match expected"
                    f" argument count {metadata.nargs}"
                )
        raise ValueError(
            f"no type in {options} could be instantiated from"
            f" {strings}.\n\nGot errors:  \n- " + "\n- ".join(errors)
        )

    return union_instantiator, InstantiatorMetadata(
        nargs=nargs,
        metavar=metavar,
        choices=None,
        action=None,
    )


def _instantiator_from_dict(
    typ: TypeForm,
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[Instantiator, InstantiatorMetadata]:
    key_type, val_type = get_args(typ)
    key_instantiator, key_meta = _instantiator_from_type_inner(
        key_type, type_from_typevar, allow_sequences="fixed_length", markers=markers
    )

    if _markers.UseAppendAction in markers:
        val_instantiator, val_meta = _instantiator_from_type_inner(
            val_type,
            type_from_typevar,
            allow_sequences=True,
            markers=markers - {_markers.UseAppendAction},
        )
        pair_metavar = f"{key_meta.metavar} {val_meta.metavar}"
        key_nargs = cast(int, key_meta.nargs)  # Casts needed for mypy but not pyright!
        val_nargs = val_meta.nargs
        assert isinstance(key_nargs, int)

        def append_dict_instantiator(strings: List[List[str]]) -> Any:
            out = {}
            for s in strings:
                out[key_instantiator(s[:key_nargs])] = val_instantiator(s[key_nargs:])  # type: ignore
            return out

        return append_dict_instantiator, InstantiatorMetadata(
            nargs=key_nargs + val_nargs if isinstance(val_nargs, int) else "*",
            metavar=pair_metavar,
            choices=None,
            action="append",
        )
    else:
        val_instantiator, val_meta = _instantiator_from_type_inner(
            val_type, type_from_typevar, allow_sequences="fixed_length", markers=markers
        )
        pair_metavar = f"{key_meta.metavar} {val_meta.metavar}"
        key_nargs = cast(int, key_meta.nargs)  # Casts needed for mypy but not pyright!
        val_nargs = cast(int, val_meta.nargs)
        assert isinstance(key_nargs, int)
        assert isinstance(val_nargs, int)
        pair_nargs = key_nargs + val_nargs

        def dict_instantiator(strings: List[str]) -> Any:
            out = {}
            if len(strings) % pair_nargs != 0:
                raise ValueError("incomplete set of key value pairs!")

            index = 0
            for _ in range(len(strings) // pair_nargs):
                assert isinstance(key_nargs, int)
                assert isinstance(val_nargs, int)
                k = strings[index : index + key_nargs]
                index += key_nargs
                v = strings[index : index + val_nargs]
                index += val_nargs

                if key_meta.choices is not None and any(
                    kj not in key_meta.choices for kj in k
                ):
                    raise ValueError(
                        f"invalid choice: {k} (choose from {key_meta.choices}))"
                    )
                if val_meta.choices is not None and any(
                    vj not in val_meta.choices for vj in v
                ):
                    raise ValueError(
                        f"invalid choice: {v} (choose from {val_meta.choices}))"
                    )
                out[key_instantiator(k)] = val_instantiator(v)  # type: ignore
            return out

        return dict_instantiator, InstantiatorMetadata(
            nargs="*",
            metavar=_strings.multi_metavar_from_single(pair_metavar),
            choices=None,
            action=None,
        )


def _instantiator_from_sequence(
    typ: TypeForm,
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[Instantiator, InstantiatorMetadata]:
    """Instantiator for variable-length sequences: list, sets, Tuple[T, ...], etc."""
    container_type = get_origin(typ)
    assert container_type is not None
    if container_type is collections.abc.Sequence:
        container_type = list

    if container_type is tuple:
        (contained_type, ell) = get_args(typ)
        assert ell == Ellipsis
    else:
        (contained_type,) = get_args(typ)

    if _markers.UseAppendAction in markers:
        make, inner_meta = _instantiator_from_type_inner(
            contained_type,
            type_from_typevar,
            allow_sequences=True,
            markers=markers - {_markers.UseAppendAction},
        )

        def append_sequence_instantiator(strings: List[List[str]]) -> Any:
            assert strings is not None
            return container_type(cast(_StandardInstantiator, make)(s) for s in strings)

        return append_sequence_instantiator, InstantiatorMetadata(
            nargs=inner_meta.nargs,
            metavar=inner_meta.metavar,
            choices=inner_meta.choices,
            action="append",
        )
    else:
        make, inner_meta = _instantiator_from_type_inner(
            contained_type,
            type_from_typevar,
            allow_sequences="fixed_length",
            markers=markers,
        )

        def sequence_instantiator(strings: List[str]) -> Any:
            # Validate nargs.
            if (
                isinstance(inner_meta.nargs, int)
                and len(strings) % inner_meta.nargs != 0
            ):
                raise ValueError(
                    f"input {strings} is of length {len(strings)}, which is not"
                    f" divisible by {inner_meta.nargs}."
                )

            # Make tuple.
            out = []
            step = inner_meta.nargs if isinstance(inner_meta.nargs, int) else 1
            for i in range(0, len(strings), step):
                out.append(make(strings[i : i + inner_meta.nargs]))  # type: ignore
            assert container_type is not None
            return container_type(out)

        return sequence_instantiator, InstantiatorMetadata(
            nargs="*",
            metavar=_strings.multi_metavar_from_single(inner_meta.metavar),
            choices=inner_meta.choices,
            action=None,
        )


def _instantiator_from_literal(
    typ: TypeForm,
    type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    markers: FrozenSet[_markers.Marker],
) -> Tuple[_StandardInstantiator, InstantiatorMetadata]:
    choices = get_args(typ)
    str_choices = tuple(x.name if isinstance(x, enum.Enum) else str(x) for x in choices)
    return (
        # Note that if string is not in str_choices, it will be caught from setting
        # `choices` below.
        lambda strings: choices[str_choices.index(strings[0])],
        InstantiatorMetadata(
            nargs=1,
            metavar="{" + ",".join(str_choices) + "}",
            choices=set(str_choices),
            action=None,
        ),
    )



from typing import Mapping, TypeVar, Union

from typing_extensions import Annotated

from .._typing import TypeForm
from ..conf import subcommand

T = TypeVar("T")


def subcommand_type_from_defaults(
    defaults: Mapping[str, T],
    descriptions: Mapping[str, str] = {},
    *,
    prefix_names: bool = True,
) -> TypeForm[T]:
    """Construct a Union type for defining subcommands that choose between defaults.

    For example, when `defaults` is set to:

    ```python
    {
        "small": Config(...),
        "big": Config(...),
    }
    ```

    We return:

    ```python
    Union[
        Annotated[
            Config,
            tyro.conf.subcommand("small", default=Config(...))
        ],
        Annotated[
            Config,
            tyro.conf.subcommand("big", default=Config(...))
        ]
    ]
    ```

    Direct use of `typing.Union` and :func:`tyro.conf.subcommand()` should generally be
    preferred, but this function can be helpful for succinctness.

    .. warning::
        The type returned by this function can be safely used as an input to
        :func:`tyro.cli()`, but for static analysis when used for annotations we
        recommend applying a `TYPE_CHECKING` guard:

        .. code-block:: python

            from typing import TYPE_CHECKING

            if TYPE_CHECKING:
                # Static type seen by language servers, type checkers, etc.
                SelectableConfig = Config
            else:
                # Runtime type used by tyro.
                SelectableConfig = subcommand_type_from_defaults(...)

    Args:
        defaults: A dictionary of default subcommand instances.
        descriptions: A dictionary conttaining descriptions for helptext.
        prefix_names: Whether to prefix subcommand names.

    Returns:
        A subcommand type, which can be passed to :func:`tyro.cli`.
    """
    return Union.__getitem__(  # type: ignore
        tuple(
            Annotated.__class_getitem__(  # type: ignore
                (
                    type(v),
                    subcommand(
                        k,
                        default=v,
                        description=descriptions.get(k, ""),
                        prefix_name=prefix_names,
                    ),
                )
            )
            for k, v in defaults.items()
        )
    )



"""Abstractions for pulling out 'field' definitions, which specify inputs, types, and # type: ignore
defaults, from general callables."""
from __future__ import annotations

import collections
import collections.abc
import dataclasses
import enum
import functools
import inspect
import itertools
import numbers
import os
import sys
import typing
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    FrozenSet,
    Hashable,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import docstring_parser
import typing_extensions
from typing_extensions import (
    Annotated,
    NotRequired,
    Required,
    get_args,
    get_origin,
    is_typeddict,
)

from . import (
    _docstrings,
    _instantiators,
    _resolver,
    _singleton,
    _strings,
    _unsafe_cache,
    conf,  # Avoid circular import.
)
from ._typing import TypeForm
from .conf import _confstruct, _markers


@dataclasses.dataclass(frozen=True)
class FieldDefinition:
    intern_name: str
    extern_name: str
    type_or_callable: Union[TypeForm[Any], Callable]
    """Type or callable for this field. This should have all Annotated[] annotations
    stripped."""
    default: Any
    # We need to record whether defaults are from default instances to
    # determine if they should override the default in
    # tyro.conf.subcommand(default=...).
    is_default_from_default_instance: bool
    helptext: Optional[str]
    markers: FrozenSet[Any]
    custom_constructor: bool

    argconf: _confstruct._ArgConfiguration

    # Override the name in our kwargs. Useful whenever the user-facing argument name
    # doesn't match the keyword expected by our callable.
    call_argname: Any

    def __post_init__(self):
        if (
            _markers.Fixed in self.markers or _markers.Suppress in self.markers
        ) and self.default in MISSING_SINGLETONS:
            raise _instantiators.UnsupportedTypeAnnotationError(
                f"Field {self.intern_name} is missing a default value!"
            )

    @staticmethod
    def make(
        name: str,
        type_or_callable: Union[TypeForm[Any], Callable],
        default: Any,
        is_default_from_default_instance: bool,
        helptext: Optional[str],
        call_argname_override: Optional[Any] = None,
        *,
        markers: Tuple[_markers.Marker, ...] = (),
    ):
        # Try to extract argconf overrides from type.
        _, argconfs = _resolver.unwrap_annotated(
            type_or_callable, _confstruct._ArgConfiguration
        )
        argconf = _confstruct._ArgConfiguration(
            None,
            None,
            help=None,
            aliases=None,
            prefix_name=True,
            constructor_factory=None,
        )
        for overwrite_argconf in argconfs:
            # Apply any annotated argument configuration values.
            argconf = dataclasses.replace(
                argconf,
                **{
                    field.name: getattr(overwrite_argconf, field.name)
                    for field in dataclasses.fields(overwrite_argconf)
                    if getattr(overwrite_argconf, field.name) is not None
                },
            )
            if argconf.help is not None:
                helptext = argconf.help

        type_or_callable, inferred_markers = _resolver.unwrap_annotated(
            type_or_callable, _markers._Marker
        )
        return FieldDefinition(
            intern_name=name,
            extern_name=name if argconf.name is None else argconf.name,
            type_or_callable=type_or_callable
            if argconf.constructor_factory is None
            else argconf.constructor_factory(),
            default=default,
            is_default_from_default_instance=is_default_from_default_instance,
            helptext=helptext,
            markers=frozenset(inferred_markers).union(markers),
            custom_constructor=argconf.constructor_factory is not None,
            argconf=argconf,
            call_argname=call_argname_override
            if call_argname_override is not None
            else name,
        )

    def add_markers(self, markers: Tuple[Any, ...]) -> FieldDefinition:
        return dataclasses.replace(
            self,
            markers=self.markers.union(markers),
        )

    def is_positional(self) -> bool:
        """Returns True if the argument should be positional in the commandline."""
        return (
            # Explicit positionals.
            _markers.Positional in self.markers
            # Dummy dataclasses should have a single positional field.
            or self.intern_name == _strings.dummy_field_name
            or (
                # Make required arguments positional.
                _markers.PositionalRequiredArgs in self.markers
                and self.default in MISSING_SINGLETONS
            )
        )

    def is_positional_call(self) -> bool:
        """Returns True if the argument should be positional in underlying Python call."""
        return (
            # Explicit positionals.
            _markers._PositionalCall in self.markers
            # Dummy dataclasses should have a single positional field.
            or self.intern_name == _strings.dummy_field_name
        )


class PropagatingMissingType(_singleton.Singleton):
    pass


class NonpropagatingMissingType(_singleton.Singleton):
    pass


class ExcludeFromCallType(_singleton.Singleton):
    pass


class NotRequiredButWeDontKnowTheValueType(_singleton.Singleton):
    pass


# We have two types of missing sentinels: a propagating missing value, which when set as
# a default will set all child values of nested structures as missing as well, and a
# nonpropagating missing sentinel, which does not override child defaults.
MISSING_PROP = PropagatingMissingType()
MISSING_NONPROP = NonpropagatingMissingType()

# When total=False in a TypedDict, we exclude fields from the constructor by default.
NOT_REQUIRED_BUT_WE_DONT_KNOW_THE_VALUE = NotRequiredButWeDontKnowTheValueType()


EXCLUDE_FROM_CALL = ExcludeFromCallType()

# Note that our "public" missing API will always be the propagating missing sentinel.
MISSING: Any = MISSING_PROP
"""Sentinel value to mark fields as missing. Can be used to mark fields passed in as a
`default_instance` for `tyro.cli()` as required."""


MISSING_SINGLETONS = [
    dataclasses.MISSING,
    MISSING_PROP,
    MISSING_NONPROP,
    inspect.Parameter.empty,
]
try:
    # Undocumented feature: support omegaconf dataclasses out of the box.
    import omegaconf

    MISSING_SINGLETONS.append(omegaconf.MISSING)
except ImportError:
    pass

DEFAULT_SENTINEL_SINGLETONS = MISSING_SINGLETONS + [
    NOT_REQUIRED_BUT_WE_DONT_KNOW_THE_VALUE,
    EXCLUDE_FROM_CALL,
]


@dataclasses.dataclass(frozen=True)
class UnsupportedNestedTypeMessage:
    """Reason why a callable cannot be treated as a nested type."""

    message: str


@_unsafe_cache.unsafe_cache(maxsize=1024)
def is_nested_type(
    typ: Union[TypeForm[Any], Callable], default_instance: DefaultInstance
) -> bool:
    """Determine whether a type should be treated as a 'nested type', where a single
    type can be broken down into multiple fields (eg for nested dataclasses or
    classes).

    TODO: we should come up with a better name than 'nested type', which is a little bit
    misleading."""

    # Need to swap types.
    _, argconfs = _resolver.unwrap_annotated(
        typ, search_type=conf._confstruct._ArgConfiguration
    )
    _, subcommand_confs = _resolver.unwrap_annotated(
        typ, search_type=conf._confstruct._SubcommandConfiguration
    )
    for union_conf in argconfs + subcommand_confs:
        if union_conf.constructor_factory is not None:
            typ = union_conf.constructor_factory()
    return not isinstance(
        _try_field_list_from_callable(typ, default_instance),
        UnsupportedNestedTypeMessage,
    )


def field_list_from_callable(
    f: Union[Callable, TypeForm[Any]],
    default_instance: DefaultInstance,
    support_single_arg_types: bool,
) -> Tuple[
    Union[Callable, TypeForm[Any]], Dict[TypeVar, TypeForm], List[FieldDefinition]
]:
    """Generate a list of generic 'field' objects corresponding to the inputs of some
    annotated callable.

    Returns:
        The type that `f` is resolved as.
        A type_from_typevar dict.
        A list of field definitions.
    """
    # Resolve generic types.
    f, type_from_typevar = _resolver.resolve_generic_types(f)
    f = _resolver.unwrap_newtype_and_narrow_subtypes(f, default_instance)

    # Try to generate field list.
    field_list = _try_field_list_from_callable(f, default_instance)

    if isinstance(field_list, UnsupportedNestedTypeMessage):
        if support_single_arg_types:
            return (
                f,
                type_from_typevar,
                [
                    FieldDefinition(
                        intern_name="value",
                        extern_name="value",  # Doesn't matter.
                        type_or_callable=f,
                        default=default_instance,
                        is_default_from_default_instance=True,
                        helptext="",
                        custom_constructor=False,
                        markers=frozenset(
                            (_markers.Positional, _markers._PositionalCall)
                        ),
                        argconf=_confstruct._ArgConfiguration(
                            None, None, None, None, None, None
                        ),
                        call_argname="",
                    )
                ],
            )
        else:
            raise _instantiators.UnsupportedTypeAnnotationError(field_list.message)

    # Recursively apply markers.
    _, parent_markers = _resolver.unwrap_annotated(f, _markers._Marker)
    field_list = list(map(lambda field: field.add_markers(parent_markers), field_list))

    # Try to resolve types in our list of fields.
    def resolve(field: FieldDefinition) -> FieldDefinition:
        typ = field.type_or_callable
        typ = _resolver.apply_type_from_typevar(typ, type_from_typevar)
        typ = _resolver.type_from_typevar_constraints(typ)
        typ = _resolver.narrow_collection_types(typ, field.default)
        typ = _resolver.narrow_union_type(typ, field.default)

        # Check that the default value matches the final resolved type.
        # There's some similar Union-specific logic for this in narrow_union_type(). We
        # may be able to consolidate this.
        if (
            # Be relatively conservative: isinstance() can be checked on non-type
            # types (like unions in Python >=3.10), but we'll only consider single types
            # for now.
            type(typ) is type
            and not isinstance(field.default, typ)  # type: ignore
            # If a custom constructor is set, field.type_or_callable may not be
            # matched to the annotated type.
            and not field.custom_constructor
            and field.default not in DEFAULT_SENTINEL_SINGLETONS
            # The numeric tower in Python is wacky. This logic is non-critical, so
            # we'll just skip it (+the complexity) for numbers.
            and not isinstance(field.default, numbers.Number)
        ):
            # If the default value doesn't match the resolved type, we expand the
            # type. This is inspired by https://github.com/brentyi/tyro/issues/88.
            warnings.warn(
                f"The field {field.intern_name} is annotated with type {field.type_or_callable}, "
                f"but the default value {field.default} has type {type(field.default)}. "
                f"We'll try to handle this gracefully, but it may cause unexpected behavior."
            )
            typ = Union[typ, type(field.default)]  # type: ignore

        field = dataclasses.replace(field, type_or_callable=typ)
        return field

    field_list = list(map(resolve, field_list))

    return f, type_from_typevar, field_list


# Implementation details below.


DefaultInstance = Union[
    Any, PropagatingMissingType, NonpropagatingMissingType, ExcludeFromCallType
]

_known_parsable_types = set(
    filter(
        lambda x: isinstance(x, Hashable),  # type: ignore
        itertools.chain(
            __builtins__.values(),  # type: ignore
            vars(typing).values(),
            vars(typing_extensions).values(),
            vars(collections.abc).values(),
        ),
    )
)


def _try_field_list_from_callable(
    f: Union[Callable, TypeForm[Any]],
    default_instance: DefaultInstance,
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    # Check for default instances in subcommand configs. This is needed for
    # is_nested_type() when arguments are not valid without a default, and this
    # default is specified in the subcommand config.
    f, found_subcommand_configs = _resolver.unwrap_annotated(
        f, conf._confstruct._SubcommandConfiguration
    )
    if len(found_subcommand_configs) > 0:
        default_instance = found_subcommand_configs[0].default

    # Unwrap generics.
    f, type_from_typevar = _resolver.resolve_generic_types(f)
    f = _resolver.apply_type_from_typevar(f, type_from_typevar)
    f = _resolver.unwrap_newtype_and_narrow_subtypes(f, default_instance)
    f = _resolver.narrow_collection_types(f, default_instance)
    f_origin = _resolver.unwrap_origin_strip_extras(cast(TypeForm, f))

    # If `f` is a type:
    #     1. Set cls to the type.
    #     2. Consider `f` to be `cls.__init__`.
    cls: Optional[TypeForm[Any]] = None
    if inspect.isclass(f):
        cls = f
        if hasattr(cls, "__init__") and cls.__init__ is not object.__init__:
            f = cls.__init__  # type: ignore
        elif hasattr(cls, "__new__") and cls.__new__ is not object.__new__:
            f = cls.__new__
        else:
            return UnsupportedNestedTypeMessage(
                f"Cannot instantiate class {cls} with no unique __init__ or __new__"
                " method."
            )
        f_origin = cls  # type: ignore

    # Try field generation from class inputs.
    if cls is not None:
        for match, field_list_from_class in (
            (is_typeddict, _field_list_from_typeddict),
            (_resolver.is_namedtuple, _field_list_from_namedtuple),
            (_resolver.is_dataclass, _field_list_from_dataclass),
            (_is_attrs, _field_list_from_attrs),
            (_is_pydantic, _field_list_from_pydantic),
        ):
            if match(cls):
                return field_list_from_class(cls, default_instance)

    # Standard container types. These are different because they can be nested structures
    # if they contain other nested types (eg Tuple[Struct, Struct]), or treated as
    # single arguments otherwise (eg Tuple[int, int]).
    #
    # Note that f_origin will be populated if we annotate as `Tuple[..]`, and cls will
    # be populated if we annotate as just `tuple`.
    if f_origin is tuple or cls is tuple:
        return _field_list_from_tuple(f, default_instance)
    elif f_origin in (collections.abc.Mapping, dict) or cls in (
        collections.abc.Mapping,
        dict,
    ):
        return _field_list_from_dict(f, default_instance)
    elif f_origin in (list, set, typing.Sequence) or cls in (
        list,
        set,
        typing.Sequence,
    ):
        return _field_list_from_nontuple_sequence_checked(f, default_instance)

    # General cases.
    if (
        cls is not None and cls in _known_parsable_types
    ) or _resolver.unwrap_origin_strip_extras(f) in _known_parsable_types:
        return UnsupportedNestedTypeMessage(f"{f} should be parsed directly!")
    elif (
        cls is not None
        and issubclass(_resolver.unwrap_origin_strip_extras(cls), os.PathLike)
        and _instantiators.is_type_string_converter(cls)
    ):
        return UnsupportedNestedTypeMessage(
            f"PathLike {cls} should be parsed directly!"
        )
    else:
        return _try_field_list_from_general_callable(f, cls, default_instance)


def _field_list_from_typeddict(
    cls: TypeForm[Any], default_instance: DefaultInstance
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    field_list = []
    valid_default_instance = (
        default_instance not in MISSING_SINGLETONS
        and default_instance is not EXCLUDE_FROM_CALL
    )
    assert not valid_default_instance or isinstance(default_instance, dict)
    total = getattr(cls, "__total__", True)
    assert isinstance(total, bool)
    assert not valid_default_instance or isinstance(default_instance, dict)
    for name, typ in _resolver.get_type_hints(cls, include_extras=True).items():
        typ_origin = get_origin(typ)
        is_default_from_default_instance = False
        if valid_default_instance and name in cast(dict, default_instance):
            default = cast(dict, default_instance)[name]
            is_default_from_default_instance = True
        elif typ_origin is Required and total is False:
            # Support total=False.
            default = MISSING_PROP
        elif total is False:
            # Support total=False.
            default = EXCLUDE_FROM_CALL
            if is_nested_type(typ, MISSING_NONPROP):
                # total=False behavior is unideal for nested structures.
                pass
                # raise _instantiators.UnsupportedTypeAnnotationError(
                #     "`total=False` not supported for nested structures."
                # )
        elif typ_origin is NotRequired:
            # Support typing.NotRequired[].
            default = EXCLUDE_FROM_CALL
        else:
            default = MISSING_PROP

        # Nested types need to be populated / can't be excluded from the call.
        if default is EXCLUDE_FROM_CALL and is_nested_type(typ, MISSING_NONPROP):
            default = MISSING_NONPROP

        if typ_origin in (Required, NotRequired):
            args = get_args(typ)
            assert (
                len(args) == 1
            ), "typing.Required[] and typing.NotRequired[T] require a concrete type T."
            typ = args[0]
            del args

        field_list.append(
            FieldDefinition.make(
                name=name,
                type_or_callable=typ,
                default=default,
                is_default_from_default_instance=is_default_from_default_instance,
                helptext=_docstrings.get_field_docstring(cls, name),
            )
        )
    return field_list


def _field_list_from_namedtuple(
    cls: TypeForm[Any], default_instance: DefaultInstance
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    # Handle NamedTuples.
    #
    # TODO: in terms of helptext, we currently do display the default NamedTuple
    # helptext. But we (intentionally) don't for dataclasses; this is somewhat
    # inconsistent.
    field_list = []
    field_defaults = getattr(cls, "_field_defaults")

    # Note that _field_types is removed in Python 3.9.
    for name, typ in _resolver.get_type_hints(cls, include_extras=True).items():
        # Get default, with priority for `default_instance`.
        default = field_defaults.get(name, MISSING_NONPROP)
        if hasattr(default_instance, name):
            default = getattr(default_instance, name)
        if default_instance is MISSING_PROP:
            default = MISSING_PROP

        field_list.append(
            FieldDefinition.make(
                name=name,
                type_or_callable=typ,
                default=default,
                is_default_from_default_instance=True,
                helptext=_docstrings.get_field_docstring(cls, name),
            )
        )
    return field_list


def _field_list_from_dataclass(
    cls: TypeForm[Any], default_instance: DefaultInstance
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    is_flax_module = False
    try:
        # Check if dataclass is a flax module. This is only possible if flax is already
        # loaded.
        #
        # We generally want to avoid importing flax, since it requires a lot of heavy
        # imports.
        if "flax.linen" in sys.modules.keys():
            import flax.linen

            if issubclass(cls, flax.linen.Module):
                is_flax_module = True
    except ImportError:
        pass

    # Handle dataclasses.
    field_list = []
    for dc_field in filter(lambda field: field.init, _resolver.resolved_fields(cls)):
        # For flax modules, we ignore the built-in "name" and "parent" fields.
        if is_flax_module and dc_field.name in ("name", "parent"):
            continue

        default, is_default_from_default_instance = _get_dataclass_field_default(
            dc_field, default_instance
        )

        # Try to get helptext from field metadata. This is also intended to be
        # compatible with HuggingFace-style config objects.
        helptext = dc_field.metadata.get("help", None)
        assert isinstance(helptext, (str, type(None)))

        # Try to get helptext from docstrings. Note that this can't be generated
        # dynamically.
        if helptext is None:
            helptext = _docstrings.get_field_docstring(cls, dc_field.name)

        field_list.append(
            FieldDefinition.make(
                name=dc_field.name,
                type_or_callable=dc_field.type,
                default=default,
                is_default_from_default_instance=is_default_from_default_instance,
                helptext=helptext,
            )
        )
    return field_list


# Support attrs and pydantic if they're installed.
try:
    import pydantic
except ImportError:
    if not TYPE_CHECKING:
        pydantic = None  # type: ignore

try:
    from pydantic import v1 as pydantic_v1
except ImportError:
    if not TYPE_CHECKING:
        pydantic_v1 = None  # type: ignore


def _is_pydantic(cls: TypeForm[Any]) -> bool:
    if pydantic is not None and issubclass(cls, pydantic.BaseModel):
        return True
    if pydantic_v1 is not None and issubclass(cls, pydantic_v1.BaseModel):
        return True
    return False


def _field_list_from_pydantic(
    cls: TypeForm[Any], default_instance: DefaultInstance
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    assert pydantic is not None

    # Handle pydantic models.
    field_list = []
    pydantic_version = int(getattr(pydantic, "__version__", "1.0.0").partition(".")[0])
    if pydantic_version < 2 or (
        pydantic_v1 is not None and issubclass(cls, pydantic_v1.BaseModel)
    ):
        # Pydantic 1.xx.
        # We do a conditional cast because the pydantic.v1 module won't
        # actually exist in legacy versions of pydantic.
        if TYPE_CHECKING:
            cls_cast = cast(pydantic_v1.BaseModel, cls)
        else:
            cls_cast = cls
        for pd1_field in cls_cast.__fields__.values():
            helptext = pd1_field.field_info.description
            if helptext is None:
                helptext = _docstrings.get_field_docstring(cls, pd1_field.name)

            default, is_default_from_default_instance = _get_pydantic_v1_field_default(
                pd1_field.name, pd1_field, default_instance
            )

            field_list.append(
                FieldDefinition.make(
                    name=pd1_field.name,
                    type_or_callable=pd1_field.outer_type_,
                    default=default,
                    is_default_from_default_instance=is_default_from_default_instance,
                    helptext=helptext,
                )
            )
    else:
        # Pydantic 2.xx.
        for name, pd2_field in cast(pydantic.BaseModel, cls).model_fields.items():
            helptext = pd2_field.description
            if helptext is None:
                helptext = _docstrings.get_field_docstring(cls, name)

            default, is_default_from_default_instance = _get_pydantic_v2_field_default(
                name, pd2_field, default_instance
            )

            field_list.append(
                FieldDefinition.make(
                    name=name,
                    type_or_callable=Annotated.__class_getitem__(  # type: ignore
                        (pd2_field.annotation,) + tuple(pd2_field.metadata)
                    )
                    if len(pd2_field.metadata) > 0
                    else pd2_field.annotation,
                    default=default,
                    is_default_from_default_instance=is_default_from_default_instance,
                    helptext=helptext,
                )
            )
    return field_list


try:
    import attr
except ImportError:
    attr = None  # type: ignore


def _is_attrs(cls: TypeForm[Any]) -> bool:
    return attr is not None and attr.has(cls)


def _field_list_from_attrs(
    cls: TypeForm[Any], default_instance: DefaultInstance
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    assert attr is not None

    # Handle attr classes.
    field_list = []
    for attr_field in attr.fields(cls):
        # Default handling.
        name = attr_field.name
        default = attr_field.default
        is_default_from_default_instance = False
        if default_instance not in MISSING_SINGLETONS:
            if hasattr(default_instance, name):
                default = getattr(default_instance, name)
                is_default_from_default_instance = True
            else:
                warnings.warn(
                    f"Could not find field {name} in default instance"
                    f" {default_instance}, which has"
                    f" type {type(default_instance)},",
                    stacklevel=2,
                )
        elif default is attr.NOTHING:
            default = MISSING_NONPROP
        elif isinstance(default, attr.Factory):  # type: ignore
            default = default.factory()  # type: ignore

        assert attr_field.type is not None
        field_list.append(
            FieldDefinition.make(
                name=name,
                type_or_callable=attr_field.type,
                default=default,
                is_default_from_default_instance=is_default_from_default_instance,
                helptext=_docstrings.get_field_docstring(cls, name),
            )
        )
    return field_list


def _field_list_from_tuple(
    f: Union[Callable, TypeForm[Any]], default_instance: DefaultInstance
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    # Fixed-length tuples.
    field_list: List[FieldDefinition] = []
    children = get_args(f)
    if Ellipsis in children:
        return _try_field_list_from_sequence_inner(
            next(iter(set(children) - {Ellipsis})), default_instance
        )

    # Infer more specific type when tuple annotation isn't subscripted. This generally
    # doesn't happen
    if len(children) == 0:
        if default_instance in MISSING_SINGLETONS:
            return UnsupportedNestedTypeMessage(
                "If contained types of a tuple are not specified in the annotation, a"
                " default instance must be specified."
            )
        else:
            assert isinstance(default_instance, tuple)
            children = tuple(type(x) for x in default_instance)

    if (
        default_instance in MISSING_SINGLETONS
        # EXCLUDE_FROM_CALL indicates we're inside a TypedDict, with total=False.
        or default_instance is EXCLUDE_FROM_CALL
    ):
        default_instance = (default_instance,) * len(children)

    for i, child in enumerate(children):
        default_i = default_instance[i]  # type: ignore
        field_list.append(
            FieldDefinition.make(
                # Ideally we'd have --tuple[0] instead of --tuple.0 as the command-line
                # argument, but in practice the brackets are annoying because they
                # require escaping.
                name=str(i),
                type_or_callable=child,
                default=default_i,
                is_default_from_default_instance=True,
                helptext="",
                # This should really set the positional marker, but the CLI is more
                # intuitive for mixed nested/non-nested types in tuples when we stick
                # with kwargs. Tuples are special-cased in _calling.py.
            )
        )

    contains_nested = False
    for field in field_list:
        if get_origin(field.type_or_callable) is Union:
            for option in get_args(field.type_or_callable):
                # The second argument here is the default value, which can help with
                # narrowing but is generall not necessary.
                contains_nested |= is_nested_type(option, MISSING_NONPROP)

        contains_nested |= is_nested_type(field.type_or_callable, field.default)

        if contains_nested:
            break
    if not contains_nested:
        # We could also check for variable length children, which can be populated when
        # the tuple is interpreted as a nested field but not a directly parsed one.
        return UnsupportedNestedTypeMessage(
            "Tuple does not contain any nested structures."
        )

    return field_list


def _field_list_from_nontuple_sequence_checked(
    f: Union[Callable, TypeForm[Any]], default_instance: DefaultInstance
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    """Tuples are handled differently due to variadics (tuple[T1, T2, T3, ..., Tn]),
    while list[], sequence[], set[], etc only have one typevar."""
    contained_type: Any
    if len(get_args(f)) == 0:
        # A raw collection type (list, tuple, etc) was passed in, but narrowing also failed.
        assert default_instance in MISSING_SINGLETONS, f"{default_instance} {f}"
        return UnsupportedNestedTypeMessage(
            f"Sequence type {f} needs either an explicit type or a"
            " default to infer from."
        )
    else:
        (contained_type,) = get_args(f)
    return _try_field_list_from_sequence_inner(contained_type, default_instance)


def _try_field_list_from_sequence_inner(
    contained_type: TypeForm[Any],
    default_instance: DefaultInstance,
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    # When no default instance is specified:
    #     If we have List[int] => this can be parsed as a single field.
    #     If we have List[SomeStruct] => OK.
    if default_instance in MISSING_SINGLETONS and not is_nested_type(
        contained_type, MISSING_NONPROP
    ):
        return UnsupportedNestedTypeMessage(
            f"Sequence containing type {contained_type} should be parsed directly!"
        )

    # If we have a default instance:
    #     [int, int, int] => this can be parsed as a single field.
    #     [SomeStruct, int, int] => OK.
    if isinstance(default_instance, Iterable) and all(
        [not is_nested_type(type(x), x) for x in default_instance]
    ):
        return UnsupportedNestedTypeMessage(
            f"Sequence with default {default_instance} should be parsed directly!"
        )
    if default_instance in MISSING_SINGLETONS:
        # We use the broader error type to prevent it from being caught by
        # is_possibly_nested_type(). This is for sure a bad annotation!
        raise _instantiators.UnsupportedTypeAnnotationError(
            "For variable-length sequences over nested types, we need a default value"
            " to infer length from."
        )

    field_list = []
    for i, default_i in enumerate(default_instance):  # type: ignore
        field_list.append(
            FieldDefinition.make(
                name=str(i),
                type_or_callable=contained_type,
                default=default_i,
                is_default_from_default_instance=True,
                helptext="",
            )
        )
    return field_list


def _field_list_from_dict(
    f: Union[Callable, TypeForm[Any]],
    default_instance: DefaultInstance,
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    if default_instance in MISSING_SINGLETONS or len(cast(dict, default_instance)) == 0:
        return UnsupportedNestedTypeMessage(
            "Nested dictionary structures must have non-empty default instance specified."
        )
    field_list = []
    for k, v in cast(dict, default_instance).items():
        field_list.append(
            FieldDefinition.make(
                name=str(k) if not isinstance(k, enum.Enum) else k.name,
                type_or_callable=type(v),
                default=v,
                is_default_from_default_instance=True,
                helptext=None,
                # Dictionary specific key:
                call_argname_override=k,
            )
        )
    return field_list


def _try_field_list_from_general_callable(
    f: Union[Callable, TypeForm[Any]],
    cls: Optional[TypeForm[Any]],
    default_instance: DefaultInstance,
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    # Generate field list from function signature.
    if not callable(f):
        return UnsupportedNestedTypeMessage(
            f"Cannot extract annotations from {f}, which is not a callable type."
        )
    params = list(inspect.signature(f).parameters.values())
    if cls is not None:
        # Ignore self parameter.
        params = params[1:]

    out = _field_list_from_params(f, cls, params)
    if isinstance(out, UnsupportedNestedTypeMessage):
        # Return error message.
        return out

    # If a default is provided: .
    if default_instance not in MISSING_SINGLETONS:
        for i, field in enumerate(out):
            out[i] = field.add_markers((_markers._OPTIONAL_GROUP,))

    return out


def _field_list_from_params(
    f: Union[Callable, TypeForm[Any]],
    cls: Optional[TypeForm[Any]],
    params: List[inspect.Parameter],
) -> Union[List[FieldDefinition], UnsupportedNestedTypeMessage]:
    # Unwrap functools.wraps and functools.partial.
    done = False
    while not done:
        done = True
        if hasattr(f, "__wrapped__"):
            f = f.__wrapped__  # type: ignore
            done = False
        if isinstance(f, functools.partial):
            f = f.func
            done = False

    # Sometime functools.* is applied to a class.
    if inspect.isclass(f):
        cls = f
        f = f.__init__  # type: ignore

    # Get type annotations, docstrings.
    docstring = inspect.getdoc(f)
    docstring_from_arg_name = {}
    if docstring is not None:
        for param_doc in docstring_parser.parse(docstring).params:
            docstring_from_arg_name[param_doc.arg_name] = param_doc.description
    del docstring

    # This will throw a type error for torch.device, typing.Dict, etc.
    try:
        hints = _resolver.get_type_hints(f, include_extras=True)
    except TypeError:
        return UnsupportedNestedTypeMessage(f"Could not get hints for {f}!")

    field_list = []
    for param in params:
        # Get default value.
        default = param.default

        # Get helptext from docstring.
        helptext = docstring_from_arg_name.get(param.name)
        if helptext is None and cls is not None:
            helptext = _docstrings.get_field_docstring(cls, param.name)

        if param.name not in hints:
            out = UnsupportedNestedTypeMessage(
                f"Expected fully type-annotated callable, but {f} with arguments"
                f" {tuple(map(lambda p: p.name, params))} has no annotation for"
                f" '{param.name}'."
            )
            if param.kind is param.KEYWORD_ONLY:
                # If keyword only: this can't possibly be an instantiator function
                # either, so we escalate to an error.
                raise _instantiators.UnsupportedTypeAnnotationError(out.message)
            return out

        # Set markers for positional + variadic arguments.
        markers: Tuple[Any, ...] = ()
        typ: Any = hints[param.name]
        if param.kind is inspect.Parameter.POSITIONAL_ONLY:
            markers = (_markers.Positional, _markers._PositionalCall)
        elif param.kind is inspect.Parameter.VAR_POSITIONAL:
            # Handle *args signatures.
            #
            # This will create a `--args T [T ...]` CLI argument.
            markers = (_markers._UnpackArgsCall,)
            typ = Tuple.__getitem__((typ, ...))  # type: ignore
            default = ()
        elif param.kind is inspect.Parameter.VAR_KEYWORD:
            # Handle *kwargs signatures.
            #
            # This will create a `--kwargs STR T [STR T ...]` CLI argument.
            #
            # Note that it would be straightforward to make both this and *args truly
            # positional, omitting the --args/--kwargs prefix, but we are choosing not
            # to because it would make *args and **kwargs difficult to use in
            # conjunction.
            markers = (_markers._UnpackKwargsCall,)
            typ = Dict.__getitem__((str, typ))  # type: ignore
            default = {}

        field_list.append(
            FieldDefinition.make(
                name=param.name,
                # Note that param.annotation doesn't resolve forward references.
                type_or_callable=typ,
                default=default,
                is_default_from_default_instance=False,
                helptext=helptext,
                markers=markers,
            )
        )

    return field_list


def _ensure_dataclass_instance_used_as_default_is_frozen(
    field: dataclasses.Field, default_instance: Any
) -> None:
    """Ensure that a dataclass type used directly as a default value is marked as
    frozen."""
    assert dataclasses.is_dataclass(default_instance)
    cls = type(default_instance)
    if not cls.__dataclass_params__.frozen:  # type: ignore
        warnings.warn(
            f"Mutable type {cls} is used as a default value for `{field.name}`. This is"
            " dangerous! Consider using `dataclasses.field(default_factory=...)` or"
            f" marking {cls} as frozen."
        )


def _get_dataclass_field_default(
    field: dataclasses.Field, parent_default_instance: Any
) -> Tuple[Any, bool]:
    """Helper for getting the default instance for a dataclass field."""
    # If the dataclass's parent is explicitly marked MISSING, mark this field as missing
    # as well.
    if parent_default_instance is MISSING_PROP:
        return MISSING_PROP, False

    # Try grabbing default from parent instance.
    if (
        parent_default_instance not in MISSING_SINGLETONS
        and parent_default_instance is not None
    ):
        # Populate default from some parent, eg `default=` in `tyro.cli()`.
        if hasattr(parent_default_instance, field.name):
            return getattr(parent_default_instance, field.name), True
        else:
            warnings.warn(
                f"Could not find field {field.name} in default instance"
                f" {parent_default_instance}, which has"
                f" type {type(parent_default_instance)},",
                stacklevel=2,
            )

    # Try grabbing default from dataclass field.
    if field.default not in MISSING_SINGLETONS:
        default = field.default
        # Note that dataclasses.is_dataclass() will also return true for dataclass
        # _types_, not just instances.
        if type(default) is not type and dataclasses.is_dataclass(default):
            _ensure_dataclass_instance_used_as_default_is_frozen(field, default)
        return default, False

    # Populate default from `dataclasses.field(default_factory=...)`.
    if field.default_factory is not dataclasses.MISSING and not (
        # Special case to ignore default_factory if we write:
        # `field: Dataclass = dataclasses.field(default_factory=Dataclass)`.
        #
        # In other words, treat it the same way as: `field: Dataclass`.
        #
        # The only time this matters is when we our dataclass has a `__post_init__`
        # function that mutates the dataclass. We choose here to use the default values
        # before this method is called.
        dataclasses.is_dataclass(field.type) and field.default_factory is field.type
    ):
        return field.default_factory(), False

    # Otherwise, no default. This is different from MISSING, because MISSING propagates
    # to children. We could revisit this design to make it clearer.
    return MISSING_NONPROP, False


def _get_pydantic_v1_field_default(
    name: str,
    field: pydantic_v1.fields.ModelField,
    parent_default_instance: DefaultInstance,
) -> Tuple[Any, bool]:
    """Helper for getting the default instance for a Pydantic field."""

    # Try grabbing default from parent instance.
    if (
        parent_default_instance not in MISSING_SINGLETONS
        and parent_default_instance is not None
    ):
        # Populate default from some parent, eg `default=` in `tyro.cli()`.
        if hasattr(parent_default_instance, name):
            return getattr(parent_default_instance, name), True
        else:
            warnings.warn(
                f"Could not find field {name} in default instance"
                f" {parent_default_instance}, which has"
                f" type {type(parent_default_instance)},",
                stacklevel=2,
            )

    if not field.required:
        return field.get_default(), False

    # Otherwise, no default.
    return MISSING_NONPROP, False


def _get_pydantic_v2_field_default(
    name: str,
    field: pydantic.fields.FieldInfo,
    parent_default_instance: DefaultInstance,
) -> Tuple[Any, bool]:
    """Helper for getting the default instance for a Pydantic field."""

    # Try grabbing default from parent instance.
    if (
        parent_default_instance not in MISSING_SINGLETONS
        and parent_default_instance is not None
    ):
        # Populate default from some parent, eg `default=` in `tyro.cli()`.
        if hasattr(parent_default_instance, name):
            return getattr(parent_default_instance, name), True
        else:
            warnings.warn(
                f"Could not find field {name} in default instance"
                f" {parent_default_instance}, which has"
                f" type {type(parent_default_instance)},",
                stacklevel=2,
            )

    if not field.is_required():
        return field.get_default(call_default_factory=True), False

    # Otherwise, no default.
    return MISSING_NONPROP, False



from typing import TYPE_CHECKING, Any, Callable, TypeVar

from typing_extensions import Annotated

from .. import _singleton

# Current design issue: markers are applied recursively to nested structures, but can't
# be unapplied.

# Note that all Annotated[T, None] values are just for static checkers. The real marker
# singletons are instantiated dynamically below.
#
# An alias could ideally be made, but SpecialForm aliases are not well supported by static analysis tools.

T = TypeVar("T")

Positional = Annotated[T, None]
"""A type `T` can be annotated as `Positional[T]` if we want to parse it as a positional
argument."""

PositionalRequiredArgs = Annotated[T, None]
"""Make all arguments without defaults positional."""

# Private marker. For when an argument is not only positional in the CLI, but also in
# the callable.
_PositionalCall = Annotated[T, None]

# Private markers for when arguments should be passed in via *args or **kwargs.
_UnpackArgsCall = Annotated[T, None]
_UnpackKwargsCall = Annotated[T, None]

# Private marker.
_OPTIONAL_GROUP = Annotated[T, None]

# TODO: the verb tenses here are inconsistent, naming could be revisited.
# Perhaps Suppress should be Suppressed? But SuppressedFixed would be weird.

Fixed = Annotated[T, None]
"""A type `T` can be annotated as `Fixed[T]` to prevent `tyro.cli` from parsing it; a
default value should be set instead. Note that fields with defaults that can't be parsed
will also be marked as fixed automatically."""

Suppress = Annotated[T, None]
"""A type `T` can be annotated as `Suppress[T]` to prevent `tyro.cli` from parsing it, and
to prevent it from showing up in helptext."""

SuppressFixed = Annotated[T, None]
"""Hide fields that are either manually or automatically marked as fixed."""

FlagConversionOff = Annotated[T, None]
"""Turn off flag conversion for booleans with default values. Instead, types annotated
with `bool` will expect an explicit True or False.

Can be used directly on boolean annotations, `FlagConversionOff[bool]`, or recursively
applied to nested types."""

AvoidSubcommands = Annotated[T, None]
"""Avoid creating subcommands when a default is provided for unions over nested types.
This simplifies CLI interfaces, but makes them less expressive.

Can be used directly on union types, `AvoidSubcommands[Union[...]]`, or recursively
applied to nested types."""

ConsolidateSubcommandArgs = Annotated[T, None]
"""Consolidate arguments applied to subcommands. Makes CLI less sensitive to argument
ordering, at the cost of support for optional subcommands.

By default, `tyro` will generate a traditional CLI interface where args are applied to
the directly preceding subcommand. When we have two subcommands `s1` and `s2`:
```
python x.py {--root options} s1 {--s1 options} s2 {--s2 options}
```

This can be frustrating because the resulting CLI is sensitive the exact positioning and
ordering of options.

To consolidate subcommands, we push arguments to the end, after all subcommands:
```
python x.py s1 s2 {--root, s1, and s2 options}
```

This is more robust to reordering of options, ensuring that any new options can simply
be placed at the end of the command>
"""

OmitSubcommandPrefixes = Annotated[T, None]
"""Make CLI inputs used for subcommands shorter by omitting the subcommand-specific
portion of the prefix.

If we have a structure with the field:

    cmd: Union[NestedTypeA, NestedTypeB]

By default, `--cmd.arg` may be generated as a flag for each dataclass in the union.
If subcommand prefixes are omitted, we would instead simply have `--arg`.

By default, `cmd:nested-type-a` and `cmd:nested-type-b` may be generated as subcommand.
If subcommand prefixes are omitted, we would instead simply have `nested-type-a` and
`nested-type-b`.
"""

OmitArgPrefixes = Annotated[T, None]
"""Make flags used for keyword arguments shorter by omitting prefixes.

If we have a structure with the field:

    cmd: NestedType

By default, `--cmd.arg` may be generated as a flag. If prefixes are omitted, we would
instead simply have `--arg`.
"""

UseAppendAction = Annotated[T, None]
"""Use "append" actions for variable-length arguments.

Given an annotation like `x: List[int]`, this means that `x = [0, 1, 2]` can be set via
the CLI syntax `--x 0 --x 1 --x 2` instead of the default of `--x 0 1 2`.

The resulting syntax may be more user-friendly; for `tyro`, it also enables support for
otherwise ambiguous annotations like `List[List[int]]`.

Can be applied to all variable-length sequences (`List[T]`, `Sequence[T]`,
`Tuple[T, ...]`, etc), including dictionaries without default values.
"""

CallableType = TypeVar("CallableType", bound=Callable)

# Dynamically generate marker singletons.
# These can be used one of two ways:
# - Marker[T]
# - Annotated[T, Marker]


class _Marker(_singleton.Singleton):
    def __getitem__(self, key):
        return Annotated.__class_getitem__((key, self))  # type: ignore


Marker = Any


def configure(*markers: Marker) -> Callable[[CallableType], CallableType]:
    """Decorator for applying configuration options.

    Configuration markers are implemented via `typing.Annotated` and straightforward to
    apply to types, for example:

    ```python
    field: tyro.conf.FlagConversionOff[bool]
    ```

    This decorator makes markers applicable to general functions as well:

    ```python
    # Recursively apply FlagConversionOff to all fields in `main()`.
    @tyro.conf.configure(tyro.conf.FlagConversionOff)
    def main(field: bool) -> None:
        ...
    ```

    Args:
        markers: Options to apply.
    """

    def _inner(callable: CallableType) -> CallableType:
        # We'll read from __tyro_markers__ in `_resolver.unwrap_annotated()`.
        callable.__tyro_markers__ = markers  # type: ignore
        return callable

    return _inner


if not TYPE_CHECKING:

    def _make_marker(description: str) -> _Marker:
        class _InnerMarker(_Marker):
            def __repr__(self):
                return description

        return _InnerMarker()

    _dynamic_marker_types = {}
    for k, v in dict(globals()).items():
        if v == Annotated[T, None]:
            _dynamic_marker_types[k] = _make_marker(k)
    globals().update(_dynamic_marker_types)
    del _dynamic_marker_types



from typing import Any, Callable, Dict, Optional, Sequence, TypeVar, Union, overload

from typing_extensions import Annotated

from .._cli import cli
from ..conf import subcommand

T = TypeVar("T")


@overload
def subcommand_cli_from_dict(
    subcommands: Dict[str, Callable[..., T]],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    use_underscores: bool = False,
) -> T:
    ...


# TODO: hack. We prefer the above signature, which Pyright understands, but as of 1.6.1
# mypy doesn't reason about the generics properly.
@overload
def subcommand_cli_from_dict(
    subcommands: Dict[str, Callable[..., Any]],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    use_underscores: bool = False,
) -> Any:
    ...


def subcommand_cli_from_dict(
    subcommands: Dict[str, Callable[..., Any]],
    *,
    prog: Optional[str] = None,
    description: Optional[str] = None,
    args: Optional[Sequence[str]] = None,
    use_underscores: bool = False,
) -> Any:
    """Generate a subcommand CLI from a dictionary of functions.

    For an input like:

    ```python
    tyro.extras.subcommand_cli_from_dict(
        {
            "checkout": checkout,
            "commit": commit,
        }
    )
    ```

    This is internally accomplished by generating and calling:

    ```python
    from typing import Annotated, Any, Union
    import tyro

    tyro.cli(
        Union[
            Annotated[
                Any,
                tyro.conf.subcommand(name="checkout", constructor=checkout),
            ],
            Annotated[
                Any,
                tyro.conf.subcommand(name="commit", constructor=commit),
            ],
        ]
    )
    ```

    Args:
        subcommands: Dictionary that maps the subcommand name to function to call.
        prog: The name of the program printed in helptext. Mirrors argument from
            `argparse.ArgumentParser()`.
        description: Description text for the parser, displayed when the --help flag is
            passed in. If not specified, `f`'s docstring is used. Mirrors argument from
            `argparse.ArgumentParser()`.
        args: If set, parse arguments from a sequence of strings instead of the
            commandline. Mirrors argument from `argparse.ArgumentParser.parse_args()`.
        use_underscores: If True, use underscores as a word delimeter instead of hyphens.
            This primarily impacts helptext; underscores and hyphens are treated equivalently
            when parsing happens. We default helptext to hyphens to follow the GNU style guide.
            https://www.gnu.org/software/libc/manual/html_node/Argument-Syntax.html
    """
    return cli(
        Union.__getitem__(  # type: ignore
            tuple(
                [
                    Annotated[
                        # The constructor function can return any object.
                        Any,
                        # We'll instantiate this object by invoking a subcommand with
                        # name k, via a constructor.
                        subcommand(name=k, constructor=v),
                    ]
                    for k, v in subcommands.items()
                ]
            )
        ),
        prog=prog,
        description=description,
        args=args,
        use_underscores=use_underscores,
    )



from typing import TYPE_CHECKING

from . import conf as conf
from . import extras as extras
from ._cli import cli as cli
from ._fields import MISSING as MISSING
from ._instantiators import (
    UnsupportedTypeAnnotationError as UnsupportedTypeAnnotationError,
)

# Deprecated interface.
if not TYPE_CHECKING:
    from ._deprecated import *  # noqa


# TODO: this should be synchronized automatically with the pyproject.toml.
__version__ = "0.7.1"



"""Rules for taking high-level field definitions and lowering them into inputs for
argparse's `add_argument()`."""
from __future__ import annotations

import argparse
import dataclasses
import enum
import functools
import itertools
import json
import shlex
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import rich.markup
import shtab

from . import _fields, _instantiators, _resolver, _strings
from ._typing import TypeForm
from .conf import _markers

if TYPE_CHECKING:
    cached_property = property
else:
    try:
        # Python >=3.8.
        from functools import cached_property
    except ImportError:
        # Python 3.7.
        from backports.cached_property import cached_property  # type: ignore


_T = TypeVar("_T")


# TODO: refactor!
class BooleanOptionalAction(argparse.Action):
    """Adapted from https://github.com/python/cpython/pull/27672"""

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        default: _T | str | None = None,
        type: Callable[[str], _T] | argparse.FileType | None = None,
        choices: Iterable[_T] | None = None,
        required: bool = False,
        help: str | None = None,
        metavar: str | tuple[str, ...] | None = None,
    ) -> None:
        _option_strings = []
        self._no_strings = set()
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith("--"):
                if "." not in option_string:
                    option_string = (
                        "--no" + _strings.get_delimeter() + option_string[2:]
                    )
                else:
                    # Loose heuristic for where to add the no-/no_ prefix.
                    left, _, right = option_string.rpartition(".")
                    option_string = left + ".no" + _strings.get_delimeter() + right
                self._no_strings.add(option_string)

                _option_strings.append(option_string)

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            assert option_string is not None
            setattr(namespace, self.dest, option_string not in self._no_strings)

    # Typically only supported in Python 3.10, but we backport some functionality in
    # _argparse_formatters.py
    def format_usage(self):
        return " | ".join(self.option_strings)


@dataclasses.dataclass(frozen=True)
class ArgumentDefinition:
    """Structure containing everything needed to define an argument."""

    intern_prefix: str  # True prefix. (eg for the argument's dest field)
    extern_prefix: str  # User-facing prefix.
    subcommand_prefix: str  # Prefix for nesting.
    field: _fields.FieldDefinition
    type_from_typevar: Dict[TypeVar, TypeForm[Any]]

    def add_argument(
        self, parser: Union[argparse.ArgumentParser, argparse._ArgumentGroup]
    ) -> None:
        """Add a defined argument to a parser."""

        # Get keyword arguments, with None values removed.
        kwargs = dict(self.lowered.__dict__)  # type: ignore
        kwargs.pop("instantiator")
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        name_or_flag = kwargs.pop("name_or_flag")
        if len(name_or_flag) == 0:
            name_or_flag = _strings.dummy_field_name

        # We're actually going to skip the default field: if an argument is unset, the
        # MISSING value will be detected in _calling.py and the field default will
        # directly be used. This helps reduce the likelihood of issues with converting
        # the field default to a string format, then back to the desired type.
        action = kwargs.get("action", None)
        if action != "append":
            kwargs["default"] = _fields.MISSING_NONPROP
        elif action == BooleanOptionalAction:
            pass
        else:
            kwargs["default"] = []

        # Apply overrides in our arg configuration object.
        # Note that the `name` field is applied when the field object is instantiated!
        if self.field.argconf.metavar is not None:
            kwargs["metavar"] = self.field.argconf.metavar

        # Add argument, with aliases if available.
        if self.field.argconf.aliases is not None and not self.field.is_positional():
            arg = parser.add_argument(
                name_or_flag, *self.field.argconf.aliases, **kwargs
            )
        else:
            if self.field.argconf.aliases is not None:
                import warnings

                warnings.warn(
                    f"Aliases were specified, but {name_or_flag} is positional. Aliases will be ignored."
                )
            arg = parser.add_argument(name_or_flag, **kwargs)

        # Do our best to tab complete paths.
        # There will be false positives here, but if choices is unset they should be
        # harmless.
        if "choices" not in kwargs:
            name_suggests_dir = (
                # The conditions are intended to be conservative; if a directory path is
                # registered as a normal file one that's OK, the reverse on the other
                # hand will be overly restrictive.
                self.field.intern_name.endswith("_dir")
                or self.field.intern_name.endswith("_directory")
                or self.field.intern_name.endswith("_folder")
            )
            name_suggests_path = (
                self.field.intern_name.endswith("_file")
                or self.field.intern_name.endswith("_path")
                or self.field.intern_name.endswith("_filename")
                or name_suggests_dir
            )
            complete_as_path = (
                # Catch types like Path, List[Path], Tuple[Path, ...] etc.
                "Path" in str(self.field.type_or_callable)
                # For string types, we require more evidence.
                or ("str" in str(self.field.type_or_callable) and name_suggests_path)
            )
            if complete_as_path:
                arg.complete = shtab.DIRECTORY if name_suggests_dir else shtab.FILE  # type: ignore

    @cached_property
    def lowered(self) -> LoweredArgumentDefinition:
        """Lowered argument definition, generated by applying a sequence of rules."""
        rules = (
            _rule_handle_defaults,
            _rule_handle_boolean_flags,
            _rule_recursive_instantiator_from_type,
            _rule_convert_defaults_to_strings,
            _rule_generate_helptext,
            _rule_set_name_or_flag_and_dest,
            _rule_positional_special_handling,
            _rule_static_cast_choices_to_patched_list,
        )
        return functools.reduce(
            lambda lowered, rule: rule(self, lowered),
            rules,
            LoweredArgumentDefinition(),
        )


@dataclasses.dataclass(frozen=True)
class LoweredArgumentDefinition:
    """Contains fields meant to be passed directly into argparse."""

    # Action that is called on parsed arguments. This handles conversions from strings
    # to our desired types.
    #
    # The main reason we use this instead of the standard 'type' argument is to enable
    # mixed-type tuples.
    instantiator: Optional[_instantiators.Instantiator] = None

    def is_fixed(self) -> bool:
        """If the instantiator is set to `None`, even after all argument
        transformations, it means that we don't have a valid instantiator for an
        argument. We then mark the argument as 'fixed', with a value always equal to the
        field default."""
        return self.instantiator is None

    # From here on out, all fields correspond 1:1 to inputs to argparse's
    # add_argument() method.
    name_or_flag: str = ""
    default: Optional[Any] = None
    dest: Optional[str] = None
    required: Optional[bool] = None
    action: Optional[Any] = None
    nargs: Optional[Union[int, str]] = None
    choices: Optional[Union[Set[str], List[str]]] = None
    # Note: unlike in vanilla argparse, our metavar is always a string. We handle
    # sequences, multiple arguments, etc, manually.
    metavar: Optional[str] = None
    help: Optional[str] = None


def _rule_handle_defaults(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> LoweredArgumentDefinition:
    """Set `required=False` if a default value is set."""

    # Mark lowered as required if a default is set.
    if (
        arg.field.default in _fields.MISSING_SINGLETONS
        and _markers._OPTIONAL_GROUP not in arg.field.markers
    ):
        return dataclasses.replace(lowered, default=None, required=True)

    return dataclasses.replace(lowered, default=arg.field.default)


def _rule_handle_boolean_flags(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> LoweredArgumentDefinition:
    if (
        _resolver.apply_type_from_typevar(
            arg.field.type_or_callable, arg.type_from_typevar
        )
        is not bool
    ):  # type: ignore
        return lowered

    if (
        arg.field.default in _fields.MISSING_SINGLETONS
        or arg.field.is_positional()
        or _markers.FlagConversionOff in arg.field.markers
        or _markers.Fixed in arg.field.markers
    ):
        # Treat bools as a normal parameter.
        return lowered
    elif arg.field.default in (True, False):
        # Default `False` => --flag passed in flips to `True`.
        return dataclasses.replace(
            lowered,
            action=BooleanOptionalAction,
            instantiator=lambda x: x,  # argparse will directly give us a bool!
        )

    assert False, (
        f"Expected a boolean as a default for {arg.field.intern_name}, but got"
        f" {lowered.default}."
    )


def _rule_recursive_instantiator_from_type(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> LoweredArgumentDefinition:
    """The bulkiest bit: recursively analyze the type annotation and use it to determine
    how to instantiate it given some string from the commandline.

    Important: as far as argparse is concerned, all inputs are strings.

    Conversions from strings to our desired types happen in the instantiator; this is a
    bit more flexible, and lets us handle more complex types like enums and multi-type
    tuples."""
    if _markers.Fixed in arg.field.markers:
        return dataclasses.replace(
            lowered,
            instantiator=None,
            metavar="{fixed}",
            required=False,
            default=_fields.MISSING_PROP,
        )
    if lowered.instantiator is not None:
        return lowered
    try:
        instantiator, metadata = _instantiators.instantiator_from_type(
            arg.field.type_or_callable,
            arg.type_from_typevar,
            arg.field.markers,
        )
    except _instantiators.UnsupportedTypeAnnotationError as e:
        if arg.field.default in _fields.MISSING_SINGLETONS:
            raise _instantiators.UnsupportedTypeAnnotationError(
                "Unsupported type annotation for the field"
                f" {_strings.make_field_name([arg.extern_prefix, arg.field.intern_name])}. To"
                " suppress this error, assign the field a default value."
            ) from e
        else:
            # For fields with a default, we'll get by even if there's no instantiator
            # available.
            return dataclasses.replace(
                lowered,
                metavar="{fixed}",
                required=False,
                default=_fields.MISSING_PROP,
            )

    if metadata.action == "append":

        def append_instantiator(x: Any) -> Any:
            out = instantiator(x)
            if arg.field.default in _fields.MISSING_SINGLETONS:
                return instantiator(x)

            return type(out)(arg.field.default) + out

        return dataclasses.replace(
            lowered,
            instantiator=append_instantiator,
            default=None,
            choices=metadata.choices,
            nargs=metadata.nargs,
            metavar=metadata.metavar,
            action=metadata.action,
            required=False,
        )
    else:
        return dataclasses.replace(
            lowered,
            instantiator=instantiator,
            choices=metadata.choices,
            nargs=metadata.nargs,
            metavar=metadata.metavar,
            action=metadata.action,
        )


def _rule_convert_defaults_to_strings(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> LoweredArgumentDefinition:
    """Sets all default values to strings, as required as input to our instantiator
    functions. Special-cased for enums."""

    def as_str(x: Any) -> Tuple[str, ...]:
        if isinstance(x, str):
            return (x,)
        elif isinstance(x, enum.Enum):
            return (x.name,)
        elif isinstance(x, Mapping):
            return tuple(itertools.chain(*map(as_str, itertools.chain(*x.items()))))
        elif isinstance(x, Sequence):
            return tuple(itertools.chain(*map(as_str, x)))
        else:
            return (str(x),)

    if (
        lowered.default is None
        or lowered.default in _fields.MISSING_SINGLETONS
        or lowered.action is not None
    ):
        return lowered
    else:
        return dataclasses.replace(lowered, default=as_str(lowered.default))


# This can be turned off when we don't want rich-based formatting. (notably for
# completion scripts)
#
# TODO: the global state here is gross. Should be revisited.
USE_RICH = True


# TODO: this function is also called outside of _arguments.py. Should be revisited.
def _rich_tag_if_enabled(x: str, tag: str) -> str:
    x = rich.markup.escape(_strings.strip_ansi_sequences(x))
    return x if not USE_RICH else f"[{tag}]{x}[/{tag}]"


def _rule_generate_helptext(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> LoweredArgumentDefinition:
    """Generate helptext from docstring, argument name, default values."""

    # If the suppress marker is attached, hide the argument.
    if _markers.Suppress in arg.field.markers or (
        _markers.SuppressFixed in arg.field.markers and lowered.is_fixed()
    ):
        return dataclasses.replace(lowered, help=argparse.SUPPRESS)

    help_parts = []

    primary_help = arg.field.helptext

    if primary_help is None and _markers.Positional in arg.field.markers:
        primary_help = _strings.make_field_name(
            [arg.extern_prefix, arg.field.intern_name]
        )

    if primary_help is not None and primary_help != "":
        help_parts.append(_rich_tag_if_enabled(primary_help, "helptext"))

    if not lowered.required:
        # Get the default value.
        # Note: lowered.default is the stringified version! See
        # `_rule_convert_defaults_to_strings()`.
        default = lowered.default
        if lowered.is_fixed() or lowered.action == "append":
            # Cases where we'll be missing the lowered default. Use field default instead.
            assert default in _fields.MISSING_SINGLETONS or default is None
            default = arg.field.default

        # Get the default value label.
        if arg.field.argconf.constructor_factory is not None:
            default_label = (
                str(default)
                if arg.field.type_or_callable is not json.loads
                else json.dumps(arg.field.default)
            )
        elif hasattr(default, "__iter__"):
            # For tuple types, we might have default as (0, 1, 2, 3).
            # For list types, we might have default as [0, 1, 2, 3].
            # For set types, we might have default as {0, 1, 2, 3}.
            #
            # In all cases, we want to display (default: 0 1 2 3), for consistency with
            # the format that argparse expects when we set nargs.
            assert default is not None
            default_label = " ".join(map(shlex.quote, map(str, default)))
        else:
            default_label = str(default)

        # Include default value in helptext. We intentionally don't use the % template
        # because the types of all arguments are set to strings, which will cause the
        # default to be casted to a string and introduce extra quotation marks.
        if lowered.instantiator is None:
            # Intentionally not quoted via shlex, since this can't actually be passed
            # in via the commandline.
            default_text = f"(fixed to: {default_label})"
        elif lowered.action == "append" and (
            default in _fields.MISSING_SINGLETONS or len(cast(tuple, default)) == 0
        ):
            default_text = "(repeatable)"
        elif lowered.action == "append" and len(cast(tuple, default)) > 0:
            assert default is not None  # Just for type checker.
            default_text = f"(repeatable, appends to: {default_label})"
        elif arg.field.default is _fields.EXCLUDE_FROM_CALL:
            # ^important to use arg.field.default and not the stringified default variable.
            default_text = "(unset by default)"
        elif (
            _markers._OPTIONAL_GROUP in arg.field.markers
            and default in _fields.MISSING_SINGLETONS
        ):
            # Argument in an optional group, but with no default. This is typically used
            # Note: lowered.default is the stringified version! See
            # `_rule_convert_defaults_to_strings()`.
            # when general (non-argument, non-dataclass) object arguments are given a
            # default, or when we use `tyro.conf.arg(constructor=...)`.
            #
            # There are some usage details that aren't communicated right now in the
            # helptext. For example: all arguments within an optional group without a
            # default should be passed in or none at all.
            default_text = "(optional)"
        elif _markers._OPTIONAL_GROUP in arg.field.markers:
            # Argument in an optional group, but which also has a default.
            default_text = f"(default if used: {default_label})"
        else:
            default_text = f"(default: {default_label})"

        help_parts.append(_rich_tag_if_enabled(default_text, "helptext_default"))
    else:
        help_parts.append(_rich_tag_if_enabled("(required)", "helptext_required"))

    # Note that the percent symbol needs some extra handling in argparse.
    # https://stackoverflow.com/questions/21168120/python-argparse-errors-with-in-help-string
    return dataclasses.replace(lowered, help=" ".join(help_parts).replace("%", "%%"))


def _rule_set_name_or_flag_and_dest(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> LoweredArgumentDefinition:
    name_or_flag = _strings.make_field_name(
        [arg.extern_prefix, arg.field.extern_name]
        if arg.field.argconf.prefix_name
        and _markers.OmitArgPrefixes not in arg.field.markers
        else [arg.field.extern_name]
    )

    # Prefix keyword arguments with --.
    if not arg.field.is_positional():
        name_or_flag = "--" + name_or_flag

    # Strip.
    if (
        # If OmitArgPrefixes was applied, then the subcommand prefix was already
        # stripped. :)
        _markers.OmitArgPrefixes not in arg.field.markers
        and _markers.Positional not in arg.field.markers
        and name_or_flag.startswith("--")
        and arg.subcommand_prefix != ""
    ):
        # This will run even when unused because we want the assert.
        strip_prefix = "--" + arg.subcommand_prefix + "."
        if _markers.OmitSubcommandPrefixes in arg.field.markers:
            assert name_or_flag.startswith(strip_prefix), name_or_flag
            name_or_flag = "--" + name_or_flag[len(strip_prefix) :]

    return dataclasses.replace(
        lowered,
        name_or_flag=name_or_flag,
        dest=_strings.make_field_name([arg.intern_prefix, arg.field.intern_name]),
    )


def _rule_positional_special_handling(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> LoweredArgumentDefinition:
    if not arg.field.is_positional():
        return lowered

    metavar = lowered.metavar
    if lowered.required:
        nargs = lowered.nargs
    else:
        if metavar is not None:
            metavar = "[" + metavar + "]"
        if lowered.nargs == 1:
            # Optional positional arguments. Note that this needs to be special-cased in
            # _calling.py.
            nargs = "?"
        else:
            # If lowered.nargs is either + or an int.
            nargs = "*"

    return dataclasses.replace(
        lowered,
        name_or_flag=_strings.make_field_name(
            [arg.intern_prefix, arg.field.intern_name]
        ),
        dest=None,
        required=None,  # Can't be passed in for positionals.
        metavar=metavar,
        nargs=nargs,
    )


class _PatchedList(list):
    """Custom list type, for avoiding "default not in choices" errors when the default
    is set to MISSING_NONPROP.

    This solves a choices error raised by argparse in a very specific edge case:
    literals in containers as positional arguments."""

    def __init__(self, li):
        super(_PatchedList, self).__init__(li)

    def __contains__(self, x: Any) -> bool:
        return list.__contains__(self, x) or x is _fields.MISSING_NONPROP


def _rule_static_cast_choices_to_patched_list(
    arg: ArgumentDefinition,
    lowered: LoweredArgumentDefinition,
) -> LoweredArgumentDefinition:
    return dataclasses.replace(
        lowered,
        choices=_PatchedList(lowered.choices) if lowered.choices is not None else None,
    )



from __future__ import annotations

import sys

import pytest

import tyro


@pytest.mark.skipif(
    sys.version_info >= (3, 10), reason="No error for newer versions of Python."
)
def test_new_union_error() -> None:
    """PEP 604 allows `|` to be used as a type annotation in Python >=3.10."""

    def main(x: int | str) -> None:
        ...

    with pytest.raises(TypeError) as e:
        tyro.cli(main)
    assert "You may be using a Union in the form of `X | Y`" in e.value.args[0]


@pytest.mark.skipif(
    sys.version_info >= (3, 9), reason="No error for newer versions of Python."
)
def test_new_collection_error() -> None:
    """PEP 585 allows standard collections to be used as generics in Python >=3.9."""

    def main(x: list[int]) -> None:
        ...

    with pytest.raises(TypeError) as e:
        tyro.cli(main)
    assert "You may be using a standard collection as a generic" in e.value.args[0]
